// File principale. Applicazione e finestra MDI.

#include <stdio.h>
#include <owl.h>
#include <mdi.h>
#include <string.h>
#include <commdlg.h>
#include <dir.h>
#include "classi.h"
#include "defines.h"
#include "usermsg.h"

//memoria RAM
extern unsigned char huge RAM[65536];

//file di help
char HelpFile[255];
//file .INI
char IniFile[255];

// Opzioni del compilatore (che e' scritto in C, non in C++... )
extern char op_a, op_b, op_c, op_e, op_g, op_s, op_w;

//funzione per liberare la memoria della watch table
extern "C" void delete_all_ws(); 

// Classe della finestra principale
class TMyMDIFrame : public TMDIFrame {
public:
  WORD ChildNum; //numero progressivo per le finestre senza ancora un nome
    
  // Finestre:
     
  TMessageWindow *Messages;     // Messaggi
  TProjectWindow *Project;      // Progetto
  TDumpWindow *Dump;            // Dump della RAM
  TDisassembler *Disasm;        // Disassemblato del programma oggetto
  TRegisterWindow *Registers;   // Registri del microprocessore
  TStackWindow *Stack;          // Stack (0100:01FF)
  TWatchWindow *Watch;          // Watch window per le variabili

  T6502 Processor;

  TMyMDIFrame(LPSTR ATitle, int MenuID);
  virtual PTWindowsObject CreateChild();
  virtual LPSTR GetClassName();
  virtual void GetWindowClass(WNDCLASS &);
  virtual void OpenFile(RTMessage)= [CM_FIRST + CM_MDIFILEOPEN];
  virtual void SetupWindow();

  virtual void FindFile(RTMessage) = [WM_USER + FIND_FILE];
  virtual void Refresh() = [WM_USER + REFRESH_ALL];
  virtual void EditByName(RTMessage) = [WM_USER + EDIT_BY_NAME];

  virtual void MenuFileLoadPRG(RTMessage) = [CM_FIRST + MENU_FILE_LOADPRG];
  virtual void MenuFileWritePRG(RTMessage) = [CM_FIRST + MENU_FILE_WRITEPRG];

  virtual void MenuWindowProject(RTMessage) = [CM_FIRST + CM_PROGETTO];
  virtual void MenuWindowMessages(RTMessage) = [CM_FIRST + CM_MESSAGGI];
  virtual void MenuWindowDump(RTMessage) = [CM_FIRST + MENU_WINDOW_DUMP];
  virtual void MenuWindowRegisters(RTMessage) = [CM_FIRST + MENU_WINDOW_REGISTERS];
  virtual void MenuWindowObject(RTMessage) = [CM_FIRST + MENU_WINDOW_OBJECT];
  virtual void MenuWindowStack(RTMessage) = [CM_FIRST + MENU_WINDOW_STACK];
  virtual void MenuWindowWatch(RTMessage) = [CM_FIRST + MENU_WINDOW_WATCH];

  virtual void MenuRunGo(RTMessage) = [CM_FIRST + MENU_RUN_GO];
  virtual void MenuRunStep(RTMessage) = [CM_FIRST + MENU_RUN_STEP];
  virtual void MenuRunReset(RTMessage) = [CM_FIRST + MENU_RUN_RESET];
  virtual void MenuRunIRQ(RTMessage) = [CM_FIRST + MENU_RUN_IRQ];
  virtual void MenuRunNMI(RTMessage) = [CM_FIRST + MENU_RUN_NMI];

  virtual void MenuDebugClear(RTMessage) = [CM_FIRST + MENU_DEBUG_CLEAR];
  virtual void MenuDebugAddWatch(RTMessage) = [CM_FIRST + MENU_DEBUG_ADD];

  virtual void MenuProjectNew(RTMessage) = [CM_FIRST + MENU_PROJECT_NEW];
  virtual void MenuProjectOpen(RTMessage) = [CM_FIRST + MENU_PROJECT_OPEN];
  virtual void MenuProjectClose(RTMessage) = [CM_FIRST + MENU_PROJECT_CLOSE];

  virtual void MenuOptionsAssembler(RTMessage) = [CM_FIRST + MENU_OPTIONS_ASSEMBLER];
  virtual void MenuOptionsDebugger(RTMessage) = [CM_FIRST + MENU_OPTIONS_DEBUGGER];

  virtual void MenuHelpAbout(RTMessage) = [CM_FIRST + MENU_HELP_ABOUT];
  virtual void MenuHelpContents(RTMessage) = [CM_FIRST + MENU_HELP_CONTENTS];

};


// Finestra dei messaggi
extern HWND FinestraMessaggi;

HWND FinestraPrincipale = NULL;

// Classe dell'applicazione
class TMDIApp : public TApplication {
public:
    TMDIApp(LPSTR name, HINSTANCE hInstance,
		  HINSTANCE hPrevInstance, LPSTR lpCmd,int nCmdShow);
    ~TMDIApp();
    virtual void InitMainWindow();
};



// --------------------------------------------------------------------- 

//Costruttore dell'oggetto applicazione.

TMDIApp::TMDIApp(LPSTR name, HINSTANCE hInstance,
		 HINSTANCE hPrevInstance, LPSTR lpCmd, int nCmdShow)
		: TApplication(name, hInstance, hPrevInstance, lpCmd, nCmdShow)
{
 char drive[MAXDRIVE], dir[MAXDIR];
 char nome[MAXFILE], ext[MAXEXT];

 //Costruisce i nomi dei files .INI e .HLP usando lo stesso path del .EXE
 GetModuleFileName(hInstance, HelpFile, 255);
 fnsplit(HelpFile, drive, dir, nome, ext);
 strcpy(nome, "6502SDK");
 strcpy(ext, "HLP");
 fnmerge(HelpFile, drive, dir, nome, ext);
 strcpy(ext, "INI");
 fnmerge(IniFile, drive, dir, nome, ext);

    //Legge le opzioni dal file .INI
 op_w = GetPrivateProfileInt("COMPILER", "Warnings", 1, IniFile);
 op_s = GetPrivateProfileInt("COMPILER", "SymbolTable", 0, IniFile);
 op_c = GetPrivateProfileInt("COMPILER", "CaseSensitivity", 1, IniFile);
}


// Costruisce la finestra principale (MDI) e carica il suo menu.
void TMDIApp::InitMainWindow()
{
  MainWindow = new TMyMDIFrame("6502 Software Development Kit", ID_MDIMENU);
  HAccTable = LoadAccelerators(hInstance, "ACCELERATORS_1");
}

TMDIApp::~TMDIApp()
{
 delete_all_ws();  //distrugge la watch table
}

// ---------------------------------------------------------------------

// Costruttore della finestra MDI.
// ChildNum e' il numero della prossima figlia senza nome.
TMyMDIFrame::TMyMDIFrame(LPSTR ATitle, int MenuID)
             : TMDIFrame(ATitle, MenuID)
{
  ChildNum = 1;
  ChildMenuPos = 8;  // posizione del menu delle finestre figlie MDI.

  Messages = NULL;
  Project = NULL;
  Dump = NULL;
  Registers = NULL;
  Disasm = NULL;
  Stack = NULL;
  Watch = NULL;
}

void TMyMDIFrame::SetupWindow()
{
   TMDIFrame::SetupWindow();
   Messages = new TMessageWindow(this);

   // Si, lo so, la TMyMDIFrame dovrebbe essere una unica
   // istanza. Ma chi lo sa cosa succedera' domani?
   // Non potevo mettere questa istruzione in TMDIApp::InitMainWindow()
   // perche' in quel punto MainWindow->HWindow e' ancora NULL.

   if(FinestraPrincipale == NULL)
	   FinestraPrincipale = GetApplication()->MainWindow->HWindow;
}

// Crea e visualizza una finestra figlia di editing sorgente
PTWindowsObject TMyMDIFrame::CreateChild()
{
  char nome[MAXPATH];
  TSourceEditor *newwindow;
  TWindowsObject *obj;

  wsprintf(nome, "noname%02d.in", ChildNum++);
  newwindow =  new TSourceEditor(this, "Source ", Messages, NULL);
  obj = GetModule()->MakeWindow(newwindow);
  newwindow->SetFileName(nome);
  return(obj);
}

// Nome della classe
LPSTR TMyMDIFrame::GetClassName()
{
 return("6502SDK_Frame");
}

// Assegna l'icona della finestra MDI
void TMyMDIFrame::GetWindowClass(WNDCLASS& AWndClass)
{
  TMDIFrame::GetWindowClass(AWndClass);
  AWndClass.hIcon = LoadIcon(GetApplication()->hInstance, "ICONA_APPL");
}

// Cerca la finestra dell'editor a cui corrisponde il file 'nome'
void TMyMDIFrame::FindFile(RTMessage msg)
{
 TWindowsObject *cur;
 LRESULT handle;

 cur = GetFirstChild();
 do
   {
    if (!strcmp(cur->nameOf(), "TSourceEditor") )
       {
	handle = SendMessage(cur->HWindow, WM_USER + IS_THIS_YOUR_FILE, 0, msg.LParam);
	if(handle)
          {
	   msg.Result = (LRESULT)(handle);
	   return;
          }
       }
    cur = cur->Next();
   }
 while ( cur!=GetFirstChild() );

 msg.Result = NULL;
}

// Rinfresca le finestre della RAM, dei registri, ecc...
void TMyMDIFrame::Refresh()
{
 if( Registers )
    {
     //InvalidateRect(Registers->HWindow, NULL, TRUE);
     //UpdateWindow(Registers->HWindow);
     Registers->Refresh();
    }
 if( Dump )
    {
     InvalidateRect(Dump->HWindow, NULL, TRUE);
     UpdateWindow(Dump->HWindow);
    }
 if( Disasm )
    {
     Disasm->FindPC();  //FindPC() fa anche il Refresh();
    }
 if(Stack)
    {
     InvalidateRect(Stack->HWindow, NULL, TRUE);
     UpdateWindow(Stack->HWindow);
    }
 if(Watch)
    {
     Watch->RicalcolaTutto();
     if(Watch->HWindow)
        {
	 InvalidateRect(Watch->HWindow, NULL, TRUE);
	 UpdateWindow(Watch->HWindow);
        }
    }
}

// Apre la finestra " (char *)LParam " alla linea (int)WParam
// (risponde al doppio click su un file del progetto o su un messaggio)
void TMyMDIFrame::EditByName(RTMessage msg)
{
 TWindowsObject *cur;
 LRESULT handle;

 cur = GetFirstChild();
 do
   {
    if (!strcmp(cur->nameOf(), "TSourceEditor") )
       {
	handle = SendMessage(cur->HWindow, WM_USER + IS_THIS_YOUR_FILE, 0, msg.LParam);
	if(handle)
          {
	   SendMessage(ClientWnd->HWindow, WM_MDIACTIVATE,(WPARAM)cur->HWindow, 0L);
	   SendMessage(cur->HWindow, WM_USER + GOTO_LINE, msg.WParam, 0L);
	   return;  //Ha trovato la finestra -> esce.
          }
       }
    cur = cur->Next();
   }
 while ( cur!=GetFirstChild() );

 //Non ha trovato la finestra? Allora la crea.
 cur = GetApplication()->MakeWindow(new TSourceEditor(this,
       "Source ", Messages, (char *)msg.LParam ) );
 SendMessage(cur->HWindow, WM_USER + GOTO_LINE, msg.WParam, 0L);
}


// Respond to "Open" command by constructing, creating, and setting up a
// new TFileWindow MDI child

void TMyMDIFrame::OpenFile(RTMessage)
{
  char ChildName[14];

  static OPENFILENAME ofn;
  char szFile[256], szFileTitle[256];
  const char szFilter[] =  "6502 Assembly input files (*.IN)\0*.in\0All files\0*.*\0";

  //OWL usa dei dialog box 'fatti in casa' che a me fanno schifo:
  //
  //if ( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILEOPEN,
  //                           _fstrcpy(FileName, "*.in"))) == IDOK )
  //	GetApplication()->MakeWindow(new TSourceEditor(this, ChildName, Messages, FileName));
  //
  // io preferisco, tutto sommato, quelli di Windows:

  szFile[0] = '\0';

  /* Set all structure members to zero. */

  memset(&ofn, 0, sizeof(OPENFILENAME));

  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = HWindow;
  ofn.lpstrFilter = szFilter;
  ofn.nFilterIndex = 1;
  ofn.lpstrFile= szFile;
  ofn.nMaxFile = sizeof(szFile);
  ofn.lpstrFileTitle = szFileTitle;
  ofn.nMaxFileTitle = sizeof(szFileTitle);
  ofn.lpstrInitialDir = NULL;
  ofn.Flags = OFN_HIDEREADONLY| OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

  if (GetOpenFileName(&ofn))
     {
      wsprintf(ChildName,"Source ");
      GetApplication()->MakeWindow(new TSourceEditor(this,
		       ChildName, Messages, ofn.lpstrFile));
     }
}

//carica un'area di RAM da un file .PRG 
void TMyMDIFrame::MenuFileLoadPRG(RTMessage)
{
 FILE *fileprg;
 unsigned char loadd, hiadd;
 int c;
 unsigned int loc;
 static OPENFILENAME ofn;
 char szFile[256], szFileTitle[256];
 const char szFilter[] =  "Program files (*.PRG)\0*.prg\0";

 szFile[0] = '\0';

 /* Set all structure members to zero. */

 memset(&ofn, 0, sizeof(OPENFILENAME));

 ofn.lStructSize = sizeof(OPENFILENAME);
 ofn.hwndOwner = HWindow;
 ofn.lpstrFilter = szFilter;
 ofn.nFilterIndex = 1;
 ofn.lpstrFile= szFile;
 ofn.nMaxFile = sizeof(szFile);
 ofn.lpstrFileTitle = szFileTitle;
 ofn.nMaxFileTitle = sizeof(szFileTitle);
 ofn.lpstrInitialDir = NULL;
 ofn.Flags = OFN_HIDEREADONLY| OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

 if (GetOpenFileName(&ofn))
     {
      if ( !(fileprg = fopen(ofn.lpstrFile, "rb")) ) return;
      loadd = (unsigned char) fgetc(fileprg);
      hiadd = (unsigned char) fgetc(fileprg);
      loc = loadd+0x100*hiadd;
      while( (c=fgetc(fileprg)) != EOF)
         RAM[loc++]=c;
      fclose(fileprg);
     }
}

//salva un'area di RAM in formato PRG
void TMyMDIFrame::MenuFileWritePRG(RTMessage)
{
 unsigned int FromAdd, ToAdd;
 FILE *fileprg;
 unsigned int loc;
 char FileName[MAXPATH];
 static OPENFILENAME ofn;
 char szFileTitle[256];
 char  szFilter[] = "Program files (*.PRG)\0 *.prg\0";

 GetApplication()->ExecDialog(new TFromToDialog(this, &FromAdd, &ToAdd));

 if (FromAdd<=ToAdd)
   {
    FileName[0] = '\0';

    /* Set all structure members to zero. */
    memset(&ofn, 0, sizeof(OPENFILENAME));

    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner = HWindow;
    ofn.lpstrFilter = szFilter;
    ofn.lpstrFile= FileName;
    ofn.nMaxFile = sizeof(FileName);
    ofn.lpstrFileTitle = szFileTitle;
    ofn.nMaxFileTitle = sizeof(szFileTitle);
    ofn.lpstrInitialDir = NULL;

    ofn.Flags = OFN_SHOWHELP | OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY;

    if (GetSaveFileName(&ofn))
       {
        if ( !(fileprg = fopen(FileName, "wb")) ) return;
        fputc(LOBYTE(FromAdd), fileprg);
        fputc(HIBYTE(FromAdd), fileprg);
        for(loc = FromAdd; loc<=ToAdd; loc++)
           fputc(RAM[loc], fileprg);
        fclose(fileprg);
       }
    else
       return;
   }
}

// Lancia il programma fino al primo BRK
void TMyMDIFrame::MenuRunGo(RTMessage)
{
 MSG messaggio;
 int result;
 int counter;
 static BOOL StoppedByBreakpoint = FALSE;
 char buf[100];
 int res;
 unsigned int this_instr;

 result = 0;
 counter = Processor.NBeforeEscape;
 while(!result)
   {
    if(Disasm)
       {
        if (StoppedByBreakpoint)
	   StoppedByBreakpoint = FALSE;
        else
	   if(Disasm->IsBreakpoint(Processor.PC))
	     {
	      result = 10;
              break;
	     }
       }
    this_instr = Processor.PC; //in caso di errore ricorda
			       //l'indirizzo dell'istruzione

    result=Processor.step();   //esegue l'istruzione

    if (result == 3)
      {
       Refresh();
       wsprintf(buf, "Stack overflow in $%04X.\nContinue anyway?", this_instr);
       res = MessageBox(FinestraPrincipale, buf, "Stack warning", MB_YESNO | MB_ICONEXCLAMATION );
       result = (res==IDNO)? 5 : 0;
      }
    else
         if(result == 4) 
           {
            Refresh();
            wsprintf(buf, "Stack underflow in $%04X.\nContinue anyway?", this_instr);
            res = MessageBox(FinestraPrincipale, buf, "Stack warning", MB_YESNO | MB_ICONEXCLAMATION );
            result = (res==IDNO)? 5 : 0;
           }
      // provvisoriamente si puo' fermare un RUN con un click del mouse
    if(!counter--)
      if( PeekMessage(&messaggio, HWindow, WM_MOUSEFIRST, WM_MOUSELAST, PM_REMOVE) )
	 if (messaggio.message == WM_LBUTTONDOWN)
            {
             Refresh();
	     counter = Processor.NBeforeEscape;
             result = 5;
	     return;
	    }
    // non faccio il Refresh() durante l'esecuzione veloce, ma solo alla fine.
   }

 Refresh();
 switch ( result )
 {
  case 1:     //istruzione BRK
      if (Processor.StopOnBreak)
           MessageBox(HWindow, "BRK reached", "Breakpoint", MB_OK);
      break;
  case 2:     //opcode errato
      MessageBox(HWindow, "Invalid opcode", "Error", MB_OK);
      break;
  case 3:     //stack overflow
      break;
  case 4:     //stack underflow
      break;
  case 5:     //interruzione utente
      break;
  case 10:    //breakpoint
      SendMessage(ClientWnd->HWindow, WM_MDIACTIVATE,(WPARAM)Disasm->HWindow, 0L);
      StoppedByBreakpoint = TRUE;
      break;
 }
}

// Esegue un passo
void TMyMDIFrame::MenuRunStep(RTMessage)
{
 switch ( Processor.step() )
 {
  case 1:
      if (Processor.StopOnBreak)
	    MessageBox(HWindow, "BRK reached", "Breakpoint", MB_OK);
      break;
  case 2:
      MessageBox(HWindow, "Invalid opcode", "Error", MB_OK);
      break;
 }
 Refresh();
}

//Resetta il processore
void TMyMDIFrame::MenuRunReset(RTMessage msg)
{
 Processor.reset();
 if(!Processor.StopAfterReset)
      MenuRunGo(msg);
 else
      Refresh();
}

void TMyMDIFrame::MenuRunIRQ(RTMessage)
{
 Processor.IRQ();
}

void TMyMDIFrame::MenuRunNMI(RTMessage)
{
 Processor.NMI();
}

//Cancella tutta la memoria
void TMyMDIFrame::MenuDebugClear(RTMessage)
{
 long i;

 for(i=0; i<=0xFFFF; i++)
 RAM[i] = 0;
 Refresh();
}

void TMyMDIFrame::MenuDebugAddWatch(RTMessage msg)
{
 if(!Watch)
    {
     Watch = new TWatchWindow(this, &Watch);
     GetApplication()->MakeWindow(Watch);
    }
 if(Watch)
    { 
     Watch->Show(SW_SHOW);
     SendMessage(ClientWnd->HWindow, WM_MDIACTIVATE,(WPARAM)Watch->HWindow, 0L);
     Watch->AddWatch(msg);
    }
}

void TMyMDIFrame::MenuWindowProject(RTMessage)
{
 if (!Project->HWindow)
	GetModule()->MakeWindow(Project);
 Project->Show(SW_SHOW);
}

void TMyMDIFrame::MenuWindowMessages(RTMessage)
{
 if(!Messages->HWindow)
      GetModule()->MakeWindow(Messages);
 Messages->Show(SW_SHOW);
 FinestraMessaggi = Messages->HWindow;
}

void TMyMDIFrame::MenuWindowDump(RTMessage)
{
 if (!Dump)
    {
     Dump = new TDumpWindow(this, &Dump);
     GetApplication()->MakeWindow(Dump);
    }
}

void TMyMDIFrame::MenuWindowObject(RTMessage)
{
 if (!Disasm)
    {
     Disasm = new TDisassembler(this, &Disasm, &Processor);
     GetApplication()->MakeWindow(Disasm);
    }
}

void TMyMDIFrame::MenuWindowRegisters(RTMessage)
{
 if (!Registers)
    {
     Registers = new TRegisterWindow(this, &Registers, &Processor);
     GetApplication()->MakeWindow(Registers);
    }
}

void TMyMDIFrame::MenuWindowStack(RTMessage)
{
 if (!Stack)
    {
     Stack = new TStackWindow(this, &Stack, &Processor);
     GetApplication()->MakeWindow(Stack);
    }
}

void TMyMDIFrame::MenuWindowWatch(RTMessage)
{
 if (!Watch)
    {
     Watch = new TWatchWindow(this, &Watch);
     GetApplication()->MakeWindow(Watch);
    }
}


//Crea un nuovo progetto
void TMyMDIFrame::MenuProjectNew(RTMessage)
{
 if(Project)
    if(Project->CloseProject()) return;
 Project = new TProjectWindow(this, &Project, Messages);
 GetApplication()->MakeWindow(Project);
}

//Apre un progetto
void TMyMDIFrame::MenuProjectOpen(RTMessage)
{
 if(Project)
    if(Project->CloseProject()) return;
 Project = new TProjectWindow(this, &Project, Messages);
 Project->OpenProject();
 GetApplication()->MakeWindow(Project);
}

//Chiude un progetto
void TMyMDIFrame::MenuProjectClose(RTMessage)
{
 if(Project)
    Project->CloseProject();
}

// Cambia le opzioni del compilatore
void TMyMDIFrame::MenuOptionsAssembler(RTMessage)
{
 GetApplication()->ExecDialog(new TOptionsWindow(this) );

}
// Cambia le opzioni del debugger
void TMyMDIFrame::MenuOptionsDebugger(RTMessage)
{
 GetApplication()->ExecDialog(new TOptDebugWindow(this, &Processor) );
}

// AboutBox
void TMyMDIFrame::MenuHelpAbout(RTMessage)
{
 GetApplication()->ExecDialog(new TDialog(this, ID_ABOUTBOX));
}

//Indice dell'Help
void TMyMDIFrame::MenuHelpContents(RTMessage)
{
 WinHelp(HWindow, HelpFile, HELP_CONTENTS, 0L); 
}


// ------------------------------------------------------------------------

// Lancia l'applicazione
int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
		   LPSTR lpCmd, int nCmdShow)
{
 TMDIApp MDIApp ("6502SDK", hInstance, hPrevInstance,
		lpCmd, nCmdShow);

 MDIApp.Run();
 return (MDIApp.Status);
}
