%{
	#pragma warn -rch
	#pragma warn -pro

	#include "multiio.h"
	#include <stdlib.h>
	#include <malloc.h>
	#include <string.h>

	#include "global.h"

	#define BRANCH 0x0100

	extern FILE *yyin;

	unsigned int loc_counter=0;     /* location counter */
	char org_direct=FALSE; /* TRUE se esiste una ORG */
	  /* spazio per le stringhe temporanee */
	char temp_string[400];

	int passo;           /* passo di assemblaggio */
	char genera;     /* e' ora di generare il codice? */
	char all_solved=TRUE;  /* TRUE se i riferimenti sono tutti risolti */
	char one_solved; /* TRUE se nell'ultimo paso e' stato risolto almeno un riferimento */

	char asm_error; /* #errori fatali */
	char asm_warning; /* #warning */
	char enable_warning=TRUE; /*TRUE: conta i warning; FALSE: non li conta */

	char last_unrs[200];    /* last unresolved reference */
	unsigned int last_unrl; /* last unres. ref. line */

        char org_link_flag;      /* TRUE se l'ORG segue dal file precedente del progetto */
	unsigned int org_value;  /* la prima direttiva ORG incontrata nell'assem-
                                    blaggio. Serve per le opzioni del debugger.*/ 

	extern char CurrentMacroName[255];
        unsigned int CurrentMacroInstance; /* Indirizzo dell'istanza della macro che si sta espandendo */

	extern void assembla(int, unsigned int, char);
	void io_message_nl_here();
%}


%union {
	 unsigned int mnem;
	 unsigned int intero;
	 char *simbolo;
	 struct t_operando {
			    unsigned int valore; /* valore numerico */
			    char calcolato;      /* TRUE o FALSE   */
			    char flag8_16;       /* TRUE: 8bit     */
			   }  operando;
       }

%token A X Y

%token  ADC AND ASL BCC BCS BEQ BIT BMI BNE BPL BRK BVC BVS CLC CLD
%token  CLI CLV CMP CPX CPY DEC DEX DEY EOR INC INX INY JMP JSR LDA
%token  LDX LDY LSR NOP ORA PHA PHP PLA PLP ROL ROR RTI RTS SBC SEC
%token  SED SEI STA STX STY TAX TAY TSX TXA TXS TYA

%token ORG
%token DB DW DD
%token EQU

%token END_OF_LINE
%token COMMENTO
%token <intero> INTERO8 INTERO16
%token <simbolo> SIMBOLO STRINGA

%type <mnem> mnem
%type <operando> operando e

%left OR_OP
%left AND_OP

%left '+' '-'
%left '*' '/'
%left uminus
%left '<' '>'
%left NOT_OP

%start init

%%
init: {
       static unsigned int firstorg;
       org_direct  = org_link_flag ;
       if (org_link_flag)
          {
	   if (passo)
	       loc_counter = firstorg;
	   else
	       firstorg = loc_counter;
          }
       else
         loc_counter = DEFAULT_START;

      } s

s: s END_OF_LINE linea
 | s END_OF_LINE
 | linea
 |
 ;


linea: label i
     | no_label i
     | label direttiva
     | direttiva_EQU
     | no_label direttiva
     | label
     ;

label: label1 {
	       /* Una vecchia versione accettava le label senza i ':'
		  sparando un warning. :

				       if(enable_warning)
				       {
					...
					asm_warning++;  invece di asm_error++
				       }
	      */

		      printfile();
		      printline();
		      io_message("Error: missing \":\" after label.");
		      io_message_nl_here();
		      asm_error++;
	       }
     | label1 ':'
     ;

label1: SIMBOLO {
		 unsigned int dummy;
                 char macroid[255];

		 if(CurrentMacroName[0])
                   {
		     /*Siamo dentro ad una macro->le label sono locali */
		    strcpy(macroid, $1);
		    strcat(macroid, "@");
		    strcat(macroid, CurrentMacroName);
		    strcat(macroid, ":");
		    sprintf(macroid+strlen(macroid), "%04X\0", CurrentMacroInstance);
		    if(!passo)
		      {
		       if(!get_symbol(macroid,&dummy))
                         {
			  add_symbol(macroid,loc_counter);
                          add_watch_symbol(macroid, loc_counter);
                         }
		       else
		      	  {
		     	   printfile();
			   printline();
			   io_message("Duplicate label : %s in macro %s.",$1,CurrentMacroName);
			   io_message_nl_here();
			   asm_error++;
			  }
		      }
		    }

                 else  /*Siamo fuori dalle macro */
		    if(!passo) 
		      {
		       if(!get_symbol($1,&dummy))
                         {
			  add_symbol($1,loc_counter);
                          add_watch_symbol($1, loc_counter);
                         }
		       else
		      	  {
		     	   printfile();
			   printline();
			   io_message("Duplicate label : %s .",$1);
			   io_message_nl_here();
			   asm_error++;
			  }
		      }
		}

no_label: {}

i: mnem '#'operando     {
			     /* immediato */
				 if($3.flag8_16)
				   assembla($1,$3.valore,0);
				 else
				   {
				    if ($3.valore > 0xFF)
				       {
					 printfile();
					 printline();
					 io_message("Value out of range");
					 io_message_nl_here();
					 asm_error++;
				       }
				    else
				       {
					if(enable_warning)
					     {
					      printfile();
					      printline();
					      io_message("Warning: INT16 converted to INT8");
					      io_message_nl_here();
					      asm_warning++;
					     }
					assembla($1,$3.valore,0);
				       }
				   }

			}

 | mnem operando        {  /* assoluto (o relativo per i branch) */
				   /* oppure pagina zero */
			       if ($1 & BRANCH)
				     {
				      assembla($1 & 0xFF,$2.valore,12);
				     }
				else
				  {
				   if($2.flag8_16)
				       assembla($1,$2.valore,2);
				   else
				      assembla($1,$2.valore,1);
				  }
			}
 | mnem A               {  /* accumulatore */
			  assembla($1,0,3);
			}
 | mnem                 {  /* implicito */
			  assembla($1,0,4);
			}
 | mnem '(' operando ',' X ')'  {  /* (puntatore,X) */
				  assembla($1,$3.valore,5);
				}
 | mnem '(' operando ')' ',' Y  {  /* (puntatore),Y */
				  assembla($1,$3.valore,6);
				}
 | mnem operando ',' X  {  /* pagina_zero,X  -  assoluto,X */
				if($2.flag8_16)
				     assembla($1,$2.valore,7);
				 else
				     assembla($1,$2.valore,9);
			}
 | mnem operando ',' Y  {  /* pagina zero,Y  -  assoluto,Y */
				 if($2.flag8_16)
				     assembla($1,$2.valore,8);
				 else
				     assembla($1,$2.valore,10);

			}
 | mnem '(' operando ')'{  /* indiretto (per JMP) */
			  assembla($1,$3.valore,11)
			}
 | error END_OF_LINE {
		      printfile();
		      printline();
		      io_message("Bad mnemonic.");
		      io_message_nl_here();
		      asm_error++;
		      yyerrok;
		     }
 ;

mnem:   ADC {$$=0;}
      | AND {$$=1;}
      | ASL {$$=2;}
      | BCC {$$=3 | BRANCH ;}
      | BCS {$$=4 | BRANCH ;}
      | BEQ {$$=5 | BRANCH ;}
      | BIT {$$=6;}
      | BMI {$$=7 | BRANCH ;}
      | BNE {$$=8 | BRANCH ;}
      | BPL {$$=9 | BRANCH ;}

      | BRK {$$=10;}
      | BVC {$$=11 | BRANCH ;}
      | BVS {$$=12 | BRANCH ;}
      | CLC {$$=13;}
      | CLD {$$=14;}
      | CLI {$$=15;}
      | CLV {$$=16;}
      | CMP {$$=17;}
      | CPX {$$=18;}
      | CPY {$$=19;}

      | DEC {$$=20;}
      | DEX {$$=21;}
      | DEY {$$=22;}
      | EOR {$$=23;}
      | INC {$$=24;}
      | INX {$$=25;}
      | INY {$$=26;}
      | JMP {$$=27;}
      | JSR {$$=28;}
      | LDA {$$=29;}

      | LDX {$$=30;}
      | LDY {$$=31;}
      | LSR {$$=32;}
      | NOP {$$=33;}
      | ORA {$$=34;}
      | PHA {$$=35;}
      | PHP {$$=36;}
      | PLA {$$=37;}
      | PLP {$$=38;}
      | ROL {$$=39;}

      | ROR {$$=40;}
      | RTI {$$=41;}
      | RTS {$$=42;}
      | SBC {$$=43;}
      | SEC {$$=44;}
      | SED {$$=45;}
      | SEI {$$=46;}
      | STA {$$=47;}
      | STX {$$=48;}
      | STY {$$=49;}

      | TAX {$$=50;}
      | TAY {$$=51;}
      | TSX {$$=52;}
      | TXA {$$=53;}
      | TXS {$$=54;}
      | TYA {$$=55;}
      ;

operando: INTERO8  {
		    $$.valore = $1;
		    $$.calcolato = TRUE;
		    $$.flag8_16 = TRUE;
		   }
	| INTERO16 {
		    $$.valore = $1;
		    $$.calcolato = TRUE;
		    $$.flag8_16 = FALSE;
		   }
	| SIMBOLO {
		   unsigned int v;
		   char macroid[255];
                   char local_symbol;  /* TRUE o FALSE */

		   local_symbol = (CurrentMacroName[0]!=0);

                   if (local_symbol)
		      {  /* Dentro una macro : prima cerca "label@nomemacro";
		            se non lo trova cerca "label".     */
		       strcpy(macroid, $1);
		       strcat(macroid, "@");
		       strcat(macroid, CurrentMacroName);
		       strcat(macroid, ":");
		       sprintf(macroid+strlen(macroid), "%04X\0",CurrentMacroInstance);
		       $$.calcolato=get_symbol(macroid,&v);
		       $$.valore = v;
		       if ($$.calcolato)
		              $$.flag8_16=(v<=0xFF);
		       else
		           {
			    if(!passo || (get_symbol_i(macroid)) )
			        $$.flag8_16=FALSE;   /* se sconosciuto e' sempre a 16 bit */
			    else
				{
				  /* in questo modo si esegue il codice
                                     per i simboli esterni :   */
                                 local_symbol = FALSE;
				}
			    }

		      }
                                /* notare: non uso 'else'  */
		   if(!local_symbol) /*Fuori dalle macro */
                      {
		       $$.calcolato=get_symbol($1,&v);
		       $$.valore = v;
		       if ($$.calcolato)
		              $$.flag8_16=(v<=0xFF);
		       else
		           {
			    if(!passo || (get_symbol_i($1)) )
			        $$.flag8_16=FALSE;   /* se sconosciuto e' sempre a 16 bit */
			    else
				{
				 printfile();
				 printline();
				 io_message("Undefined symbol '%s'",$1);
				 io_message_nl_here();
				 add_symbol_i($1);
				 one_solved=TRUE;/* per non avere 'circular refer.' */
				 asm_error++;
				}
			    }
		       }

		  }
	| e '+' e  {
		    $$.calcolato=$1.calcolato && $3.calcolato;
		    $$.valore = $1.valore+$3.valore;
		    if ($$.calcolato)
			$$.flag8_16 = ($$.valore < 0x0100);
		    else
		        $$.flag8_16 = FALSE;
		   }
	| e '-' e  {
		    $$.calcolato=$1.calcolato && $3.calcolato;
		    $$.valore = $1.valore-$3.valore;
		    if ($$.calcolato)
			$$.flag8_16 = ($$.valore < 0x0100);
		    else
		        $$.flag8_16 = FALSE;
		   }
	| e '*' e  {
		    $$.calcolato=$1.calcolato && $3.calcolato;
		    $$.valore = $1.valore*$3.valore;
		    if ($$.calcolato)
			$$.flag8_16 = ($$.valore < 0x0100);
		    else
		        $$.flag8_16 = FALSE;
		   }
	| e '/' e  {
		    $$.calcolato=$1.calcolato && $3.calcolato;
		    $$.valore = $1.valore/$3.valore;
		    if ($$.calcolato)
			$$.flag8_16 = ($$.valore < 0x0100);
		    else
		        $$.flag8_16 = FALSE;
		   }

	| e OR_OP e   {
		       $$.calcolato=$1.calcolato && $3.calcolato;
		       $$.valore = $1.valore|$3.valore;
		       $$.flag8_16 = ($1.flag8_16 && $3.flag8_16);
		      }
	| e AND_OP e  {
		       $$.calcolato=$1.calcolato && $3.calcolato;
		       $$.valore = $1.valore&$3.valore;
		       $$.flag8_16 = ($1.flag8_16 && $3.flag8_16);
		      }
	| NOT_OP e    {
		       $$.calcolato=$2.calcolato;
		       $$.valore =  $2.valore^0xFFFF;
		       $$.flag8_16 = ($2.flag8_16);
		      }
	| '>' e       {
		       $$.calcolato=($2.calcolato);
		       $$.valore =  ($2.valore&0xFF00)/0x0100;
		       $$.flag8_16 = TRUE;
		      }
	| '<' e       {
		       $$.calcolato=($2.calcolato);
		       $$.valore =  ($2.valore&0x00FF);
		       $$.flag8_16 = TRUE;
		      }
	;

direttiva: ORG operando {
			 if ($2.calcolato)
			  {
			   loc_counter=$2.valore;
			   if(passo) check_memory();
			   if(!org_direct)
			      org_value = $2.valore;
			      org_direct = TRUE;
			  }
			 else
			  {
			   printfile();
			   printline();
			   io_message("ORG cannot be forward-referenced.");
			   io_message_nl_here();
			   asm_error++;
			  }
			}
	 | DB operandiB {}
	 | DW operandiW  {}
	 | DD operandiD {}
	 ;

operandiB: opB      {}
	 | operandiB ',' opB   {}
	 ;

operandiW: opW                  {}
	 | operandiW ',' opW     {}
	 ;

operandiD: opD                  {}
	 | operandiD ',' opD     {}
	 ;

opB:   operando	  {
		   if ($1.flag8_16==FALSE)
			{
			 if(enable_warning)
			      {
			       printfile();
			       printline();
			       io_message("Warning: 16-bit DB divided into hi e lo byte");
			       io_message_nl_here();
			       asm_warning++;
			      }
			 a_byte(($1.valore & 0xFF00)/0xFF);
			 a_byte(($1.valore & 0x00FF));
			}
		   else
			{
			 a_byte($1.valore);
			}
		  }
  | STRINGA  {
	      int i;
	      for (i=0; i<strlen($1)-1; i++)
		  a_byte($1[i], TRUE);
	     }
  ;

opW: operando {
		a_byte(($1.valore & 0x00FF),$1.calcolato);
		a_byte(($1.valore & 0xFF00)/0xFF,$1.calcolato);
	       }

opD: operando {
		a_byte(($1.valore & 0xFF00)/0xFF,$1.calcolato);
		a_byte(($1.valore & 0x00FF),$1.calcolato);
	       }

eqs: EQU | '=';  /*  a EQU n  si puo' scrivere come   a = n  */

direttiva_EQU: SIMBOLO eqs {strcpy(temp_string,$1); }
			   operando {    unsigned int v;

					 if(!get_symbol(temp_string,&v))
					   {
					    if($4.calcolato)
					       {
						add_symbol(temp_string,$4.valore);
                                                add_watch_symbol(temp_string, $4.valore);
						one_solved = TRUE;
					       }
					    else
						{
						 all_solved = FALSE;
						 if((passo==0) && (!get_symbol_i(temp_string)) )
						      add_symbol_i(temp_string);
						 last_unrl = line_number[include_stack_ptr];
						 strcpy(last_unrs, temp_string);
						}
					   }
					 else
					   {
					    if(enable_warning && $4.calcolato && ( $4.valore !=v) )
						{
						 printfile();
						 printline();
						 io_message("Warning: identifier \"%s\" redefined.",temp_string);
						 io_message_nl_here();
						 asm_warning++;
						}
					   }
				    }

/* sottoespressione */
e : operando      {
		   $$.valore = $1.valore;
		   $$.calcolato=$1.calcolato;
		   $$.flag8_16=$1.flag8_16;
		  }
  | '(' operando ')'    {
			  $$.valore = $2.valore;
			  $$.calcolato=$2.calcolato;
			  $$.flag8_16=$2.flag8_16;
			}
  ;

%%

#pragma warn +rch


int yyerror()
{
 return(0);
}

//manda un messaggio riferito al file e alla riga corrente
void io_message_nl_here()
{
 io_message_nl(include_name[include_stack_ptr], line_number[include_stack_ptr]);
}
