// TDisassembler: finestra con il disassemblato
// del programma oggetto. Somiglia alla TDumpWindow.

#include "disasm.h"

#include <stdio.h>
#include <string.h>

#define max(a,b)(((a) >(b)) ?(a) :(b))
#define min(a,b)(((a) <(b)) ?(a) :(b))
#define abs(a)  (((a) >= 0)  ?(a) :-(a))

#define TEXT_HEIGHT 20
#define LEFT_TEXT 20

// memoria RAM 
extern unsigned char huge RAM[65536];
// decodifica opcode
extern "C" char decodifica(unsigned char, char *, char *);
// mnemonici degli opcodes(eh, eh...)
static const char op_string[] =
 "ADCANDASLBCCBCSBEQBITBMIBNEBPLBRKBVCBVSCLCCLD\
CLICLVCMPCPXCPYDECDEXDEYEORINCINXINYJMPJSRLDA\
LDXLDYLSRNOPORAPHAPHPPLAPLPROLRORRTIRTSSBCSEC\
SEDSEISTASTXSTYTAXTAYTSXTXATXSTYA";

//-----------------Scroller verticale speciale --------------------

TObjectVScroller::TObjectVScroller(PTDisassembler TheWindow, int TheXUnit
	 , int TheYUnit, long TheXRange, long TheYRange)
   :TScroller(TheWindow, TheXUnit, TheYUnit, TheXRange, TheYRange)
{
 Window = TheWindow;
}

// Devo cambiare i primi quattro 'case' perche' il disassembler
// funzioni correttamente

/* Scrolls vertically according to scroll action and thumb position. */
void TObjectVScroller::VScroll(WORD ScrollEvent, int ThumbPos)
{
  int delta;
  long NewYPos;
  unsigned int tempaddress;
  int nlinee;
  RECT Area;

  unsigned int addrvect[60]; // max 60 linee per pagina mi sembra ragionevole.
  int lastline;

  switch (ScrollEvent) {

    case SB_LINEDOWN : Window->Disassembla(YPos, &delta);    // in avanti e' facile.
                       NewYPos = max(0, min(YPos+delta, YRange));
                       if (NewYPos != YPos)
                           {
			    ScrollWindow(Window->HWindow, 0, -YUnit, NULL, NULL);
                            YPos = NewYPos;
                            UpdateWindow(Window->HWindow);
                           }
		       break;

 // All'indietro e' piu' difficile. Supponiamo che in 10 bytes ci stiano
 // un po' di istruzioni e che ci si riesca sempre a sincronizzare
 // ( ma sara' poi vero? In teoria no, ma in pratica pare che funzioni.
 // Sperimentalmente ho visto che 6 bytes non sempre sono sufficienti ).
    case SB_LINEUP   :
		       tempaddress=YPos-10;
		       while(tempaddress < YPos)
			   {
			    Window->Disassembla(tempaddress, &delta);
                            tempaddress+=delta;
			   }  
		       if (tempaddress != YPos) //non sincronizzato - disassembla all'indietro
			    {
			     Window->Disassembla(YPos-3, &delta); // 3bytes?
			     if (delta !=3)
				{
			         Window->Disassembla(YPos-2, &delta); // 2bytes?
				 if (delta !=2)
				     delta =1;   // un solo byte. Pazienza.
				}                             
			    }
		       NewYPos = max(0, min(YPos-delta, YRange));
		       if (NewYPos != YPos)
                           {
			    ScrollWindow(Window->HWindow, 0, YUnit, NULL, NULL);
                            YPos = NewYPos;
                            UpdateWindow(Window->HWindow);
                           }
		       break;

    case SB_PAGEDOWN : NewYPos = Window->LastValidAddress;
                       if (NewYPos != YPos)
			   {
			    YPos = NewYPos;
			    InvalidateRect(Window->HWindow, NULL, TRUE);
                            UpdateWindow(Window->HWindow);
                           }
		       break;

//Saltare una pagina indietro e pretendere che l'ultima istruzione
//diventi la prima della nuova pagina e' veramente un casino.
//Lo facciamo lo stesso, con un algoritmo di "disassemblaggio all'indietro"
//che permette di avere "probabilmente" intatta l'istruzione.

    case SB_PAGEUP   : // Calcola il numero delle linee
                       GetClientRect(Window->HWindow, &Area);
		       nlinee = (Area.bottom-Area.top)/TEXT_HEIGHT ;

// Parto 10 bytes prima dell'indirizzo di partenza di 'caso peggiore':
		       tempaddress=YPos-(nlinee*3)-10;
                       lastline = 0;
		       while(tempaddress < YPos)
			   {
			    Window->Disassembla(tempaddress, &delta);
			    addrvect[lastline++]=tempaddress;
			    tempaddress+=delta;
			   }

		       if (tempaddress != YPos)
                             //non sincronizzato. Provo a traslare di 1 byte 
			    {
			     tempaddress=addrvect[lastline-nlinee]-1;
			     while(tempaddress < YPos)
			       {
				Window->Disassembla(tempaddress, &delta);
				tempaddress+=delta;
			       }
			     if(tempaddress == YPos)
				NewYPos = max(0, min(addrvect[lastline-nlinee]-1, YRange));
			     else
				NewYPos = max(0, min(addrvect[lastline-nlinee]-2, YRange));
			    }
		       else
			  NewYPos = max(0, min(addrvect[lastline-nlinee], YRange));

		       if (NewYPos != YPos)
			   {
			    YPos = NewYPos;
			    InvalidateRect(Window->HWindow, NULL, TRUE);
                            UpdateWindow(Window->HWindow);
			   }
		       break;


    case SB_TOP      : ScrollTo(XPos, 0);
                       break;
    case SB_BOTTOM   : ScrollTo(XPos, YRange);
                       break;

    case SB_THUMBPOSITION :
      if ( YRange > MaxInt )
	ScrollTo(XPos, YRangeValue(ThumbPos));
      else
        ScrollTo(XPos, ThumbPos);
      break;

    case SB_THUMBTRACK :
      if ( TrackMode )
        if ( YRange > MaxInt )
          ScrollTo(XPos, YRangeValue(ThumbPos));
        else
          ScrollTo(XPos, ThumbPos);
      if ( Window && HasVScrollBar )
        SetScrollPos(Window->HWindow, SB_VERT, ThumbPos, TRUE);
      break;
  }  // switch
}


//---------------------Finestra-------------------------------------


TDisassembler::TDisassembler(TWindowsObject *AParent, TDisassembler **Me, T6502 *proc)
      : TWindow(AParent, "Object")
{
 Myself = Me;

 Processor = proc;
 FirstBreakpoint = NULL;
 Attr.Style |= WS_VSCROLL;
 Attr.W = 300;
 Scroller = new TObjectVScroller(this, 0, TEXT_HEIGHT, 0, 0xFFFF );
 Scroller->AutoOrg = FALSE;
 LastValidAddress = 0;
}

TDisassembler::~TDisassembler()
{
 RemoveAllBreakpoints();
 *Myself = NULL;
}

void TDisassembler::SetupWindow()
{
 TWindow::SetupWindow();
 FindPC();
}

LPSTR TDisassembler::GetClassName()
{
 return("Disassembler");
}


// Assegna l'icona della finestra
void TDisassembler::GetWindowClass(WNDCLASS& AWndClass)
{
  TWindow::GetWindowClass(AWndClass);
  AWndClass.hIcon = LoadIcon(GetApplication()->hInstance, "ICONA_DEBUGGER");
  AWndClass.style |= CS_DBLCLKS;
}

// Modifica il menu della finestra MDI
void TDisassembler::ActivationResponse(WORD Activated, BOOL IsIconified)
{
 HWND handle;
 HMENU hm;

 TWindow::ActivationResponse(Activated, IsIconified);
 handle = GetApplication()->MainWindow->HWindow;
 hm = GetMenu(handle);
 if(Activated)
     {
//menu File
      EnableMenuItem(hm, CM_FILESAVE,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_FILESAVEAS,  MF_BYCOMMAND | MF_GRAYED);
//menu Edit
      EnableMenuItem(hm, CM_EDITUNDO,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCUT,     MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCOPY,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITPASTE,   MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITDELETE,  MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCLEAR,   MF_BYCOMMAND | MF_GRAYED);
//menu Search
      EnableMenuItem(hm, CM_EDITFIND,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITREPLACE, MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITFINDNEXT,MF_BYCOMMAND | MF_GRAYED);
//menu Assembler
      EnableMenuItem(hm, MENU_ASSEMBLER_ASSEMBLE, MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, MENU_ASSEMBLER_BUILDALL, MF_BYCOMMAND | MF_GRAYED);
     }
 else
     {
//menu File
      EnableMenuItem(hm, CM_FILESAVE,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_FILESAVEAS,  MF_BYCOMMAND | MF_ENABLED);
//menu Edit
      EnableMenuItem(hm, CM_EDITUNDO,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCUT,     MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCOPY,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITPASTE,   MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITDELETE,  MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCLEAR,   MF_BYCOMMAND | MF_ENABLED);
//menu Search
      EnableMenuItem(hm, CM_EDITFIND,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITREPLACE, MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITFINDNEXT,MF_BYCOMMAND | MF_ENABLED);
//menu Assembler
      EnableMenuItem(hm, MENU_ASSEMBLER_ASSEMBLE, MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, MENU_ASSEMBLER_BUILDALL, MF_BYCOMMAND | MF_ENABLED);
     }
 DrawMenuBar(handle);
}

void TDisassembler::Paint(HDC PaintDC, PAINTSTRUCT _FAR &)
{
 int y=0;
 unsigned int displ=0;
 unsigned int disasm_add;
 HFONT hfont;
 COLORREF ColorOld, Color;
 RECT Area;
 int nbytes;
 int nlinee;
 char *pt;

 // Calcola il numero delle linee
 GetClientRect(HWindow, &Area);
 nlinee = (Area.bottom-Area.top)/TEXT_HEIGHT +1;

 // Colore per i breakpoints;
 Color = RGB(255, 0, 0) ;

 hfont = (HFONT) GetStockObject(ANSI_FIXED_FONT);
 DeleteObject(SelectObject(PaintDC, hfont));
 while ( y<nlinee )
    {
     disasm_add = Scroller->YPos+displ;
     pt = Disassembla(disasm_add, &nbytes);
     if (disasm_add==Processor->PC)
	  TextOut(PaintDC, 0, y*TEXT_HEIGHT, ">>", 2 );
     if (IsBreakpoint(disasm_add))
	  {
               // i breakpoints sono in campo rosso
	   ColorOld = SetBkColor(PaintDC, Color);
	   TextOut(PaintDC, LEFT_TEXT, y*TEXT_HEIGHT, pt, strlen(pt) );
	   Color = SetBkColor(PaintDC, ColorOld);
	  }
     else      
	  TextOut(PaintDC, LEFT_TEXT, y*TEXT_HEIGHT, pt, strlen(pt) );
     displ +=nbytes;
     y++;
    }
 LastValidAddress = disasm_add; // serve per il PageDown
}

void TDisassembler::FindPC()
{
 int distance;

 // Facciamo in modo che l'istruzione corrente sia sempre visualizzata.

 if (Processor->PC > LastValidAddress)   // vai in avanti
     {
      Disassembla(LastValidAddress, &distance);  
      if(Processor->PC == LastValidAddress + distance) // avanti di 1 
	 {
	  Disassembla(Scroller->YPos, &distance);
          Scroller->YPos +=distance;
	 }
      else
          Scroller->YPos = Processor->PC; 
     }
  else if (Processor->PC < Scroller->YPos)   // vai all'indietro
     {
      Scroller->YPos = Processor->PC;
     }

  InvalidateRect(HWindow, NULL, TRUE);
  UpdateWindow(HWindow);
}

// addr e' un breakpoint? (TRUE o FALSE).
// Questo sistema (linear search) funziona bene supponendo di avere al massimo
// un centinaio di breakpoint attivi (che comunque sono gia' tanti).

BOOL TDisassembler::IsBreakpoint(unsigned int addr)
{
 Breakpoint *pt = FirstBreakpoint;

 while(pt !=NULL)
    {
     if (pt->address == addr)
	 return (TRUE);
     pt = pt->next;
    }
 return (FALSE);
}

// Inverte lo stato di un breakpoint all'indirizzo addr.
void TDisassembler::ToggleBreakpoint(unsigned int addr)
{
 Breakpoint *pt=FirstBreakpoint;
 Breakpoint *Previous = NULL;

 while(pt !=NULL)
    {
     if(pt->address == addr)
         {
	  if(Previous == NULL)
		FirstBreakpoint = pt->next;
	  else
		Previous->next = pt->next;

          delete pt;
          return;
	 }
     else
         {
	  Previous = pt;
	  pt=pt->next;
         }
    }
 pt = new Breakpoint;
 pt->address = addr;
 pt->next = FirstBreakpoint;
 FirstBreakpoint = pt;
}

// Distrugge tutti i breakpoint
void TDisassembler::RemoveAllBreakpoints(void)
{
 Breakpoint *pt;

 while(FirstBreakpoint !=NULL)
    {
     pt = FirstBreakpoint;
     FirstBreakpoint=FirstBreakpoint->next;
     delete pt;
    }
}

//Setta un breakpoint con il doppio click:
void TDisassembler::WMLButtonDoubleClick(RTMessage msg)
{
 int linea;
 int delta;
 unsigned int add;

 linea = HIWORD(msg.LParam) / TEXT_HEIGHT ;
 add = Scroller->YPos;
 for ( ; linea>0; linea--)
    {
     Disassembla(add, &delta);
     add +=delta;
    }
 ToggleBreakpoint(add);
 InvalidateRect(HWindow, NULL, TRUE);
 UpdateWindow(HWindow);
}

// Comandi dei menu
void TDisassembler::MenuDebugToggle(RTMessage)
{
 MessageBeep(-1);
     // provvisorio
}

void TDisassembler::MenuDebugRemoveAll(RTMessage)
{
 RemoveAllBreakpoints();
}


//-------------- Codice del disassemblatore ------------------------------

// op_string8 e op_string16 restituiscono un char *
// dell'argomento
char *TDisassembler::op_string8(unsigned char n)
{
 static char buffer[8];

 sprintf(buffer, "$%02X",n);
 return(buffer);
}

char *TDisassembler::op_string16(unsigned int n)
{
 static char buffer[8];

 sprintf(buffer, "$%04X",n);
 return(buffer);
}

/*
   Routine del disassemblatore vero e proprio. 
   Restituisce il puntatore alla stringa dell'istruzione e, in *nbytes, il
   numero di bytes usati (1, 2 oppure 3)
*/
char *TDisassembler::Disassembla(unsigned int start,int *nbytes)
{
 char istr, modo, trovato;
 int n;
 char buffer[20];
 static char buffer_out[50];

 n=RAM[start];
 trovato =decodifica(n, &istr, &modo);
 if(trovato)
      {
       strncpy(buffer, (op_string+3*(--istr)), 3);     // opcode
      }
 else
      {
       strcpy(buffer,"???");
       *nbytes = 1;
      }
 buffer[3]= '\0'; // Termina la stringa

 if(trovato)
  switch(--modo)
    {
     unsigned int destination; /* per i branch */

     case 0:
	 strcat(buffer, " #");
	 strcat(buffer, op_string8(RAM[start+1]) );
	 *nbytes = 2;
	 break;

     case 1:
	 strcat(buffer, " ");
	 strcat(buffer, op_string16(RAM[start+1] + 0x0100*RAM[start+2]) );
	 *nbytes = 3;
	 break;

     case 2:
	 strcat(buffer, " ");
	 strcat(buffer, op_string8(RAM[start+1]) );
	 *nbytes = 2;
	 break;

     case 3:
	 strcat(buffer, " A");
	 *nbytes = 1;
	 break;

     case 4:
	 *nbytes = 1;
	 break;

     case 5:
	 strcat(buffer, " (");
	 strcat(buffer, op_string8(RAM[start+1]) );
	 strcat(buffer, ",X)");
	 *nbytes = 2;
	 break;

     case 6:
	 strcat(buffer, " (");
	 strcat(buffer, op_string8(RAM[start+1]) );
	 strcat(buffer, "),Y");
	 *nbytes = 2;
	 break;

     case 7:
	 strcat(buffer, " ");
	 strcat(buffer, op_string8(RAM[start+1]) );
	 strcat(buffer, ",X");
	 *nbytes = 2;
	 break;

     case 8:
	 strcat(buffer, " ");
	 strcat(buffer, op_string8(RAM[start+1]) );
	 strcat(buffer, ",Y");
	 *nbytes = 2;
	 break;

     case 9:
	 strcat(buffer, " ");
	 strcat(buffer, op_string16(RAM[start+1] + 0x0100*RAM[start+2]) );
	 strcat(buffer, ",X");
	 *nbytes = 3;
	 break;

     case 10:
	 strcat(buffer, " ");
	 strcat(buffer, op_string16(RAM[start+1] + 0x0100*RAM[start+2]) );
	 strcat(buffer, ",Y");
	 *nbytes = 3;
	 break;

     case 11:
	 strcat(buffer, " (");
	 strcat(buffer, op_string16(RAM[start+1] + 0x0100*RAM[start+2]) );
	 strcat(buffer, ")");
	 *nbytes = 3;
	 break;

     case 12:
	 strcat(buffer, " ");
	 destination = start + 2 + (signed char) RAM[start+1];
	 strcat(buffer, op_string16(destination) );
	 *nbytes = 2;
	 break;

    }

 switch(*nbytes)
 {
  case 3:
    sprintf(buffer_out, "$%04X> %02X %02X %02X  \0",
	     start, RAM[start], RAM[start+1], RAM[start+2] );
    break;

  case 2:
    sprintf(buffer_out, "$%04X> %02X %02X     \0",
             start, RAM[start], RAM[start+1] );
    break;

  default:
    sprintf(buffer_out, "$%04X> %02X        \0", start, RAM[start] );
    break;
 }
 strcat(buffer_out, buffer);
 return(buffer_out);
}

