/*
		      Gestione delle macro.
                      --------------------

La sintassi per dichiarare una macro e':

	#macro nome [(param1, param2, ..., paramN)]
	[corpo della macro]
	#endm

e per chiamarla si scrive:

	nome stringa1, stringa2, ..., stringaN

Le macro vanno dichiarate PRIMA della chiamata. La lista
delle macro e' locale al file.
L'espansione delle macro viene fata da LEX usando le
start conditions. Quando LEX incontra un identificatore mentre
sta espandendo una macro, controlla che non sia un parametro.
*/

#include <stdlib.h>
#include <string.h>

#define TRUE 1
#define FALSE 0

  /* aggiunge una macro alla lista */
void AddMacro(char *nome);

  /* cerca una macro (restituisce TRUE o FALSE) */
int GetMacro(char *nome);

  /* aggiunge il parametro 'nome_param' alla macro 'nome' */                                                    
void AddMacroParam(char *nome, char *nome_param);  

  /* setta il puntatore al testo della macro a una macro */
void SetMacroText(char *nome, char *text);

  /* restituisce il puntatore al testo della macro */
char *GetMacroText(char *nome);

  /* Setta il valore di un parametro */
void SetMacroParam(char *nome, int numero, char *text);

  /* restituisce il puntatore al valore attuale di un
     parametro, oppure NULL se il parametro non esiste */
char *GetMacroParam(char *nome, char *nome_param);

  /* distrugge tutte le macro e i puntatori per l'espansione */
void DestroyAllMacro();


/* macro e parametri per la definizione */
typedef struct t_macroparam{
			    char pname[32];
                            char value[255];
			    struct t_macroparam *next;
                           }   macroparam;

typedef struct t_macro{
		       char name[32];
		       macroparam *parlist;
                       char *text;
		       struct t_macro *next;
		      }   macro;

macro *first_macro = NULL;


/* Funzioni per le macro annidate. */
typedef struct t_mbuf{
		      char *text_pointer;
		      char macro_name[255];
		      struct t_mbuf *next;
                      unsigned int instance;
                     } mbuf;

char *macro_pointer = NULL;
char CurrentMacroName[255];
mbuf *firstmbuf = NULL;
extern unsigned int CurrentMacroInstance;

void MacroBufferPush()
{
 mbuf *newmbuf;

 newmbuf = malloc(sizeof(mbuf));
 newmbuf->next = firstmbuf;
 firstmbuf=newmbuf;
 strcpy(newmbuf->macro_name, CurrentMacroName);
 newmbuf->instance = CurrentMacroInstance;
 newmbuf->text_pointer = macro_pointer;
}

void MacroBufferPop()
{
 mbuf *oldmbuf;

 if(firstmbuf)
   {
    oldmbuf = firstmbuf;
    firstmbuf = firstmbuf->next;
    strcpy(CurrentMacroName, oldmbuf->macro_name);
    CurrentMacroInstance = oldmbuf->instance;
    macro_pointer = oldmbuf->text_pointer;
    free(oldmbuf); 
   }
}



void AddMacro(char *nome)
{
 macro *newm;

 newm = malloc(sizeof(macro));
 newm->next = first_macro;
 first_macro = newm;
 strcpy(newm->name, nome);
 newm->parlist = NULL;
 newm->text = NULL;
}

int GetMacro (char *nome)
{
 macro *pt = first_macro;

 while(pt)
    if (!strcmp(pt->name, nome) )
	return(TRUE);
    else
	pt = pt->next;
 return(FALSE);
}



void AddMacroParam(char *nome, char *nome_param)
{
 macro *pt = first_macro;
 macroparam *par;
 macroparam *newpar;

 while(pt)
    if (!strcmp(pt->name, nome) )
	break;
    else
	pt = pt->next;
 if(pt)    /* I parametri vanno messi in ordine da SX a DX  */
   {
    newpar = malloc(sizeof(macroparam));
    if(!pt->parlist)
	pt->parlist = newpar;
    else
       {
	par = pt->parlist;
	while(par->next)
  	     par = par->next;
        par->next = newpar;
       }
    newpar->next = NULL;
    newpar->value[0] = NULL;
    strcpy(newpar->pname, nome_param);
   }
}

void SetMacroText(char *nome, char *text)
{
 macro *pt = first_macro;

 while(pt)
    if (!strcmp(pt->name, nome) )
	break;
    else
	pt = pt->next;
 if(pt)
    pt->text = text;
}

char *GetMacroText(char *nome)
{
 macro *pt = first_macro;

 while(pt)
    if (!strcmp(pt->name, nome) )
	return(pt->text);
    else
	pt = pt->next;
 return(NULL);
}


void SetMacroParam(char *nome, int numero, char *text)
{
 macro *pt = first_macro;
 macroparam *par;

 while(pt)
    if (!strcmp(pt->name, nome) )
	break;
    else
	pt = pt->next;
 if(pt)
   {
    par = pt->parlist;
    while( ((numero--)!=0) && (par) )
       par = par->next;
    if(par)
       strcpy(par->value, text);
   }
}

char *GetMacroParam(char *nome, char *nome_param)
{
 macro *pt = first_macro;
 macroparam *par;

 while(pt)
    if (!strcmp(pt->name, nome) )
	break;
    else
	pt = pt->next;
 if(pt)
   {
    par = pt->parlist;
    while( (par) && strcmp(par->pname, nome_param) )
       par = par->next;
    if(par)
       return(par->value);
    else
       return(NULL);
   }
 else
   return (NULL);
}

void DestroyAllMacro()
{
 macro *ptm;
 macroparam *ptp;
 mbuf *ptb;

 while(first_macro)
   {
    ptm = first_macro;
    first_macro = first_macro->next;
    while (ptm->parlist)
	 {
	  ptp = ptm->parlist;
	  ptm->parlist = ptm->parlist->next;
          free (ptp);
	 }
    free(ptm);
   }

 while(firstmbuf)
   {
    ptb = firstmbuf;
    firstmbuf=firstmbuf->next;
    free(ptb);
   }
 macro_pointer = NULL;
 CurrentMacroName[0] = '\0';
}