#include "multiio.h"
#include <stdlib.h>
#include <string.h>
#include "global.h"

/* memoria RAM per l'assemblaggio */
extern unsigned char huge RAM[65536];

/* Passo di assemblaggio */
extern int passo;

extern int open_buffer (char *);
extern void close_buffer ();
extern void destroy_all_buffers();

/* opzioni */
char op_a=TRUE;  /* stampa gli indirizzi, se in modo testo */
char op_b=TRUE;  /* TRUE: output binario,  FALSE: output testo ASCII */
char op_c=TRUE;  /* TRUE: case sensitive per gli identificatori */
char op_e=FALSE; /* echo su schermo */
char op_g=TRUE;  /* genera il codice */
char op_m=TRUE;  /* assembla in memoria */
char op_s=FALSE; /* stampa la symbol table */
char op_w=TRUE;  /* abilita i warning */

/* flag per i warning */

char a_zero  ;   /* 'assembling in zero page' */
char a_stack ;   /* 'assembling in stack' */
char a_wrap  ;   /* 'code over $FFFF' */

extern char org_link_flag;      /* TRUE se l'ORG segue dal file precedente del progetto */

FILE *uscita;
FILE *file_symbt;

const int op_table[56][13] =
{
/* IMM  ASS   P0  ACC  IMP  pX   pY   p0X  p0Y assX assY indir  rel.  */

{  105, 109, 101,  -1,  -1,  97, 113, 117,  -1, 125, 121,  -1,  -1 }, /* ADC */
{   41,  45,  37,  -1,  -1,  33,  49,  53,  -1,  61,  57,  -1,  -1 }, /* AND */
{   -1,  14,   6,  10,  -1,  -1,  -1,  22,  -1,  61,  -1,  -1,  -1 }, /* ASL */
{   -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, 144 }, /* BCC */
{   -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, 176 }, /* BCS */

{   -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, 240 }, /* BEQ */
{   -1,  44,  36,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* BIT */
{   -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  48 }, /* BMI */
{   -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, 208 }, /* BNE */
{   -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  16 }, /* BPL */

{   -1,  -1,  -1,  -1,   0,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* BRK */
{   -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  80 }, /* BVC */
{   -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, 112 }, /* BVS */
{   -1,  -1,  -1,  -1,  24,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* CLC */
{   -1,  -1,  -1,  -1, 216,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* CLD */

/* IMM  ASS   P0  ACC  IMP  pX   pY   p0X  p0Y assX assY indir  rel.  */

{   -1,  -1,  -1,  -1,  88,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* CLI */
{   -1,  -1,  -1,  -1, 184,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* CLV */
{  201, 205, 197,  -1,  -1, 193, 209, 213,  -1, 221, 217,  -1,  -1 }, /* CMP */
{  224, 236, 228,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* CPX */
{  192, 204, 196,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* CPY */

{   -1, 206, 198,  -1,  -1,  -1,  -1, 214,  -1, 222,  -1,  -1,  -1 }, /* DEC */
{   -1,  -1,  -1,  -1, 202,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* DEX */
{   -1,  -1,  -1,  -1, 136,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* DEY */
{   73,  77,  69,  -1,  -1,  65,  81,  85,  -1,  93,  89,  -1,  -1 }, /* EOR */
{   -1, 238, 230,  -1,  -1,  -1,  -1, 246,  -1, 254,  -1,  -1,  -1 }, /* INC */

{   -1,  -1,  -1,  -1, 232,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* INX */
{   -1,  -1,  -1,  -1, 200,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* INY */
{   -1,  76,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1, 108,  -1 }, /* JMP */
{   -1,  32,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* JSR */
{  169, 173, 165,  -1,  -1, 161, 177, 181,  -1, 189, 185,  -1,  -1 }, /* LDA */

/* IMM  ASS   P0  ACC  IMP  pX   pY   p0X  p0Y assX assY indir  rel.  */

{  162, 174, 166,  -1,  -1,  -1,  -1,  -1, 182,  -1, 190,  -1,  -1 }, /* LDX */
{  160, 172, 164,  -1,  -1,  -1,  -1,  -1,  -1, 188,  -1,  -1,  -1 }, /* LDY */
{   -1,  78,  70,  74,  -1,  -1,  -1,  86,  -1,  94,  -1,  -1,  -1 }, /* LSR */
{   -1,  -1,  -1,  -1, 234,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* NOP */
{    9,  13,   5,  -1,  -1,   1,  17,  21,  -1,  29,  25,  -1,  -1 }, /* ORA */

{   -1,  -1,  -1,  -1,  72,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* PHA */
{   -1,  -1,  -1,  -1,   8,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* PHP */
{   -1,  -1,  -1,  -1, 104,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* PLA */
{   -1,  -1,  -1,  -1,  40,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* PLP */
{   -1,  46,  38,  42,  -1,  -1,  -1,  54,  -1,  62,  -1,  -1,  -1 }, /* ROL */

{   -1, 110, 102, 106,  -1,  -1,  -1, 118,  -1, 126,  -1,  -1,  -1 }, /* ROR */
{   -1,  -1,  -1,  -1,  64,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* RTI */
{   -1,  -1,  -1,  -1,  96,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* RTS */
{  233, 237, 229,  -1,  -1, 225, 241, 245,  -1, 253, 249,  -1,  -1 }, /* SBC */
{   -1,  -1,  -1,  -1,  56,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* SEC */

/* IMM  ASS   P0  ACC  IMP  pX   pY   p0X  p0Y assX assY indir  rel.  */

{   -1,  -1,  -1,  -1, 248,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* SED */
{   -1,  -1,  -1,  -1, 120,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* SEI */
{   -1, 141, 133,  -1,  -1, 129, 145, 149,  -1, 157, 153,  -1,  -1 }, /* STA */
{   -1, 142, 134,  -1,  -1,  -1,  -1,  -1, 150,  -1,  -1,  -1,  -1 }, /* STX */
{   -1, 140, 132,  -1,  -1,  -1,  -1, 148,  -1,  -1,  -1,  -1,  -1 }, /* STY */

{   -1,  -1,  -1,  -1, 170,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* TAX */
{   -1,  -1,  -1,  -1, 168,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* TAY */
{   -1,  -1,  -1,  -1, 186,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* TSX */
{   -1,  -1,  -1,  -1, 138,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* TXA */
{   -1,  -1,  -1,  -1, 154,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }, /* TXS */
{   -1,  -1,  -1,  -1, 152,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1 }  /* TYA */
};

/* Scrive 'val' in 'addr' (dopo aver, eventualmente, controllato che non sia ROM...) */
void asm_poke(unsigned int addr, unsigned char val)
{
 /* provvisoriamente supponiamo chi ci siano 64k di sola RAM... */

 RAM[addr] = val;
}

/* decodifica l'istruzione 'opcode' - restituisce TRUE se e' un opcode valido*/
char decodifica (unsigned char opcode, char *istr, char *modo)
{
  /* versione "stupida" :
  ****************************************************
  int trovato = FALSE;

  for(*istr=0;(*istr<56 && !trovato);(*istr)++)
      for(*modo=0; (*modo<13 && !trovato); (*modo)++)
	   if(op_table[*istr][*modo]==opcode)
	       trovato = TRUE;
  return(trovato);
  ****************************************************

  la versione 'intelligente' costruisce, la prima volta, una
  tabella 'inversa' da usare per la decodifica :      */

  static char tab_flag = TRUE;
  static int tabella_inversa[256][2];
  int i,m,n,val;

  if (tab_flag)
     {
      for(i=0; i<256; i++)
	     tabella_inversa[i][0] = -1;
      for(i=0; i<56; i++)
          for(m=0; m<13; m++)
	     {
	      n = op_table[i][m];
              if(n>=0)
                 {
	          tabella_inversa[n][0] = i;
		  tabella_inversa[n][1] = m;
                 }
             }
      tab_flag = FALSE;
     }
  *istr = val = tabella_inversa[opcode][0];
  *modo = tabella_inversa[opcode][1];

  //provvisorio
  (*istr)++;
  (*modo)++;

  return(val != -1);
}

/* assembla l'istruzione 'mnem' */
void assembla(int mnem, unsigned int operando, char modo)
{
 int opcode_i;
 unsigned char opcode;
 unsigned int valore;
 int displacement;

 unsigned int location;

 location = loc_counter;
 opcode_i=op_table[mnem][modo];

 if (opcode_i<0)
	 {
	  /* prima di sparare un errore, controlla che non si possa
	     tentare un indirizzamento alternativo */
	  switch(modo)
	   {
	    case 2:
	       modo = 1; break;
	    case 7:
	       modo = 9; break;
	    case 8:
	       modo=10; break;
	   }
	  opcode_i=op_table[mnem][modo];
	  if(opcode_i<0)
	     {
	      printfile();
	      printline();
	      io_message("Wrong addressing mode.");
	      io_message_nl_here();
	      asm_error++;
	     }
	 }
 opcode=(unsigned char)opcode_i;
 switch(modo)
	    {
	     case 3:
	     case 4:
		if(genera)
		 {
		  if (op_e)
		      io_echo(">%04X: %02X\n", loc_counter, opcode);
		  if (op_m)
		      asm_poke(loc_counter, opcode);
		  if (op_b)
		      fprintf(uscita,"%c", (unsigned char) opcode);
		  else
		  {
		   if(op_a) fprintf(uscita,">%04X: ",loc_counter);
		   fprintf(uscita,"%02X\n",  opcode);
		  }
		 }
		loc_counter++;
		break;
	     case 0:
	     case 2:
	     case 5:
	     case 6:
	     case 7:
	     case 8:
	       valore = operando;
	       if(genera)
		   {
		    if(op_e)
		      {
		       io_echo(">%04X: %02X", loc_counter, opcode);
		       io_echo(" %02X\n", valore & 0xFF );
		      }
		    if(op_m)
		      {
		       asm_poke(loc_counter, opcode);
		       asm_poke(loc_counter+1, valore & 0xFF);
		      }
		    if (op_b)
		       fprintf(uscita,"%c%c", opcode, valore & 0xFF);
		    else
		      {
		       if(op_a) fprintf(uscita,">%04X: ",loc_counter);
		       fprintf(uscita,"%02X", opcode);
		       fprintf(uscita," %02X\n", valore & 0xFF );
		      }
		   }
		loc_counter+=2;
		break;
	    case 1:
	    case 9:
	    case 10:
	    case 11:
	       valore = operando;
	       if(genera)
		   {
		    if(op_e)
		      {
		       io_echo(">%04X: %02X", loc_counter, opcode);
		       io_echo(" %02X", valore & 0xFF);
		       io_echo(" %02X\n",(valore&0xFF00)/0x0100);
		      }
		    if(op_m)
		      {
		       asm_poke(loc_counter, opcode);
		       asm_poke(loc_counter+1, valore & 0xFF);
		       asm_poke(loc_counter+2, (valore&0xFF00)/0x0100 );
		      }
		    if (op_b)
		       fprintf(uscita,"%c%c%c", opcode, valore & 0xFF,(valore&0xFF00)/0x0100 );
		    else
		      {
		       if(op_a) fprintf(uscita,">%04X: ",loc_counter);
		       fprintf(uscita,"%02X", opcode);
		       fprintf(uscita," %02X", valore & 0xFF);
		       fprintf(uscita," %02X\n",(valore&0xFF00)/0x0100);
		      }
		   }
		loc_counter+=3;
		break;
	    case 12:
	       if(genera)
		   {
		    location = loc_counter+2;
		    displacement = operando-location;
		    valore = (signed char) displacement;
		    if (valore != displacement)
		      {
                       printfile();
		       printline();
		       io_message("Branch out of range.");
		       io_message_nl_here();
		       asm_error++;
		      }
		    if(op_e)
		      {
		       io_echo(">%04X: %02X", loc_counter, opcode);
		       io_echo(" %02X\n", valore & 0xFF);
		      }
		    if(op_m)
		      {
		       asm_poke(loc_counter, opcode);
		       asm_poke(loc_counter+1, valore & 0xFF);
		      }
		    if (op_b)
		       fprintf(uscita,"%c%c", opcode, valore & 0xFF );
		    else
		      {
		       if(op_a) fprintf(uscita,">%04X: ",loc_counter);
		       fprintf(uscita,"%02X", opcode);
		       fprintf(uscita," %02X\n", valore & 0xFF);
		      }
		   }
	       loc_counter +=2;
	       break;
	     }

 if( (op_w) && (location>loc_counter) )
     warning_wrap();
}

void a_byte(unsigned char valore)
{
 unsigned int location = loc_counter;
 /*scrive un byte (per DB e DW)*/
 if(genera)
      {
       if(op_e)
	   io_echo(">%04X: %02X\n",loc_counter, valore);
       if(op_m)
	   asm_poke(loc_counter, valore);
       if(op_b)
	   fprintf(uscita,"%c", valore);
       else
	  {
	   if(op_a) fprintf(uscita,">%04X: ",loc_counter);
	   fprintf(uscita," %02X\n", valore);
	  }
      }
 loc_counter++;
 if( op_w && (location>loc_counter) )
     warning_wrap();
}

void check_memory(void) /* controlla di non assemblare codice in pagina zero o
			   nello stack */
{
 if(op_w)
     {
       if( (loc_counter < 0x100) && (!a_zero) )
		  {
		   a_zero = TRUE;
		   io_message("Warning: assembling code in zero page ($0000:$00FF) may cause problems.");
		   io_message_nl(NULL, 0);
		   asm_warning++;
		  }
       if( (loc_counter>=0x0100) && (loc_counter<0x0200) && (!a_stack) )
		  {
		   a_stack = TRUE;
		   io_message("Warning: assembling code in stack area ($0100:$01FF) may cause problems.");
		   io_message_nl(NULL, 0);
		   asm_warning++;
		  }
     }
}

void warning_wrap()
{
  if(!a_wrap)
     {
      a_wrap = TRUE;
      io_message("Warning: code exceeds $FFFF, restarting from $0000.");
      io_message_nl(NULL, 0);
      asm_warning++;
     }
}

/* stampa il numero di linea per i messaggi di errore */
void printline(void)
{
 io_message("line %u: ",line_number[include_stack_ptr]);
}

/* stampa il nome del file per i messaggi di errore */
void printfile(void)
{
 io_message("\"%s\" - ",include_name[include_stack_ptr]);
}

void winasm(char *nome_in, char *nome_out)
{
 extern char genera;
 extern FILE *yyin;
 extern char one_solved;
 extern char all_solved;

 io_message("Input file:  %s",nome_in);
 io_message_nl(nome_in, 0);
 io_message("Output file: %s",nome_out);
 io_message_nl(nome_out, 0);

 asm_error=0;
 asm_warning=0;

 a_zero = a_stack  = a_wrap = FALSE;

 genera = FALSE;
 enable_warning=FALSE;
 passo = 0;
 include_name[0]=nome_in;
 while(!genera && !asm_error)
   {
     io_status("File %s: pass %d...\n", nome_in, passo+1);
     if (open_buffer(nome_in) == NULL)
	 {
	  io_message("File ""%s"" not found.",nome_in);
	  io_message_nl(NULL, 0);
	  destroy_all_buffers();
	  return;
	 }
/*     yyin = ingresso;  */
     all_solved=TRUE;
     one_solved=FALSE;
     line_number[include_stack_ptr]=0;
     yyparse();
     passo++;
     genera = all_solved;
     if (!one_solved && (!all_solved))
	   {
	    printfile();
	    io_message("line %4u: ",last_unrl);
	    io_message("cannot solve circular references for '%s'.",last_unrs);
	    io_message_nl_here();
	    asm_error++;
	    break;
	   }
     close_buffer();
   }
 if(op_g && !asm_error)
   {
    io_status("File %s: pass %d...\n", nome_in, passo+1);
    /* ingresso = ... */
    open_buffer(nome_in);
    if(op_b)
	uscita = fopen(nome_out,"wb");
    else
	uscita = fopen(nome_out,"wt");
/*    yyin=ingresso; */
    line_number[include_stack_ptr]=0;
//    asm_warning=0;
    enable_warning = op_w;
    yyparse();
    fclose(uscita);
    if(!org_direct && op_w && !org_link_flag)
      {
       io_message("Warning: no ORG specified, code assembled from $%X.",DEFAULT_START);
       io_message_nl(NULL, 0);
       asm_warning++;
      }
    close_buffer();
   }

 if(op_s)
  {
   file_symbt = fopen("FILE.SYM","wt");
   gen_symbol_table(file_symbt);
   fclose(file_symbt);
   if(op_e)  print_symbol_table();
  }

 delete_symbol_table();
 delete_symbol_table_i();

 if(!asm_error)
    io_message("No errors - ");
 else
   {
    io_message("%d errors - ",asm_error);
    unlink(nome_out);
   }
 if(!asm_warning)
    io_message("no warnings.");
 else
    io_message("%d warnings.",asm_warning);
 io_message_nl(NULL, 0);

 io_message_nl(NULL, 0);


 destroy_all_buffers();
}

