/* Classe TOptDebugWindow */

#include <dialog.h>
#include <stdio.h>
#include "optdebug.h"

#include "defines.h"

extern char IniFile[]; //file .INI per le opzioni

TOptDebugWindow::TOptDebugWindow(PTWindowsObject AParent, T6502 *Proc)
      : TDialog(AParent, ID_OPTION_DEBUGGER)
{
 Processore = Proc;
}

void TOptDebugWindow::SetupWindow()
{
 char buffer[8];
 TDialog::SetupWindow();
     SendDlgItemMsg(ID_OPSTARTADDRESS, EM_LIMITTEXT, (WPARAM)4, 0);

 if(Processore->OrgFromCode)
     SendDlgItemMsg(ID_OPSTARTORG, BM_SETCHECK, BF_CHECKED, 0);
 else
     SendDlgItemMsg(ID_OPSTARTDEFAULT, BM_SETCHECK, BF_CHECKED, 0);

 if(Processore->ResetToOrg)
     SendDlgItemMsg(ID_OPRESTART, BM_SETCHECK, BF_CHECKED, 0);
 else
     if(Processore->StopAfterReset)
	  SendDlgItemMsg(ID_OPRESETSTOP, BM_SETCHECK, BF_CHECKED, 0);
     else
	  SendDlgItemMsg(ID_OPRESETGO, BM_SETCHECK, BF_CHECKED, 0);

 if(Processore->CheckStackOverflow)
     SendDlgItemMsg(ID_STACK_WARNING, BM_SETCHECK, BF_CHECKED, 0);
 else
     SendDlgItemMsg(ID_NO_STACK_WARNING, BM_SETCHECK, BF_CHECKED, 0);

 SendDlgItemMsg(ID_OPCLEARFLAGS, BM_SETCHECK, Processore->ClearFlagsOnReset, 0);
 SendDlgItemMsg(ID_OPCLEARAXY, BM_SETCHECK, Processore->ClearRegsOnReset, 0);
 SendDlgItemMsg(ID_OPCLEARSP, BM_SETCHECK, Processore->ClearStackOnReset, 0);

 sprintf(buffer, "%d", Processore->NBeforeEscape);
 SendDlgItemMsg(ID_NCHECKBEFORESTOP, WM_SETTEXT, 0, (LPARAM)buffer);

 sprintf(buffer, "%04X", Processore->DefaultRestartAddress);
 SendDlgItemMsg(ID_OPSTARTADDRESS, WM_SETTEXT, 0, (LPARAM)buffer);
}

LPSTR TOptDebugWindow::GetClassName()
{
 return("OptDebug");
}

// Assegna l'icona della finestra delle opzioni
void TOptDebugWindow::GetWindowClass(WNDCLASS& AWndClass)
{
  TDialog::GetWindowClass(AWndClass);
  AWndClass.hIcon = LoadIcon(GetApplication()->hInstance, "ICONA_APPL");
}

// Cambia le opzioni
void TOptDebugWindow::WMCommand(RTMessage msg)
{
 char buf[30];

 if(msg.WP.Lo)
    // se il messaggio arriva da un controllo:
    switch(msg.WParam)
       {
	case ID_OPSTARTORG :    // restart dall'ORG del file
        case ID_OPSTARTDEFAULT: // restart dal default address
	     Processore->OrgFromCode =
		  (char)SendDlgItemMsg(ID_OPSTARTORG, BM_GETCHECK, 0, 0L);
	     sprintf(buf, "%d", Processore->OrgFromCode);
	     WritePrivateProfileString("DEBUGGER", "OrgFromCode", buf, IniFile);
	     break;
	case ID_OPRESETSTOP :    // Ferma dopo in JMP($FFFE) 
	case ID_OPRESETGO :      // Riparti  "   "     "  "
        case ID_OPRESTART :      // JMP <restart> e stop.
	     Processore->StopAfterReset =
	      !(char)SendDlgItemMsg(ID_OPRESETGO, BM_GETCHECK, 0, 0L);
	     sprintf(buf, "%d", Processore->StopAfterReset);
	     WritePrivateProfileString("DEBUGGER", "StopAfterReset", buf, IniFile);
	     Processore->ResetToOrg =
	      (char)SendDlgItemMsg(ID_OPRESTART, BM_GETCHECK, 0, 0L);
	     sprintf(buf, "%d", Processore->ResetToOrg);
	     WritePrivateProfileString("DEBUGGER", "ResetToOrg", buf, IniFile);

	case ID_STACK_WARNING :
        case ID_NO_STACK_WARNING :
	     Processore->CheckStackOverflow =
		  (char)SendDlgItemMsg(ID_STACK_WARNING, BM_GETCHECK, 0, 0L);
	     sprintf(buf, "%d", Processore->CheckStackOverflow);
	     WritePrivateProfileString("DEBUGGER", "CheckStackOverflow", buf, IniFile);
	     break;

	case ID_OPCLEARFLAGS:
	     Processore->ClearFlagsOnReset =
		  (char)SendDlgItemMsg(ID_OPCLEARFLAGS, BM_GETCHECK, 0, 0L);
	     sprintf(buf, "%d", Processore->ClearFlagsOnReset);
	     WritePrivateProfileString("DEBUGGER", "ClearFlagsOnReset", buf, IniFile);
	     break;
	case ID_OPCLEARAXY:
	     Processore->ClearRegsOnReset =
		  (char)SendDlgItemMsg(ID_OPCLEARAXY, BM_GETCHECK, 0, 0L);
	     sprintf(buf, "%d", Processore->ClearRegsOnReset);
	     WritePrivateProfileString("DEBUGGER", "ClearRegsOnReset", buf, IniFile);
	     break;
	case ID_OPCLEARSP:
	     Processore->ClearStackOnReset =
		  (char)SendDlgItemMsg(ID_OPCLEARSP, BM_GETCHECK, 0, 0L);
	     sprintf(buf, "%d", Processore->ClearStackOnReset);
	     WritePrivateProfileString("DEBUGGER", "ClearStackOnReset", buf, IniFile);
	     break;
       }
 TDialog::WMCommand(msg);
}

// Quando esce, aggiorna l'indirizzo di restart e NBeforeEscape
void TOptDebugWindow::CloseWindow(int n)
{
 char buffer[10];

 SendDlgItemMsg(ID_OPSTARTADDRESS, WM_GETTEXT, 9, (LPARAM)buffer);
 Processore->DefaultRestartAddress = (unsigned int)strtol(buffer, NULL, 16);
 sprintf(buffer, "0x%04X", Processore->DefaultRestartAddress);
 WritePrivateProfileString("DEBUGGER", "DefaultRestartAddress", buffer, IniFile);

 SendDlgItemMsg(ID_NCHECKBEFORESTOP, WM_GETTEXT, 9, (LPARAM)buffer);
 Processore->NBeforeEscape = (unsigned int)strtol(buffer, NULL, 10);
 sprintf(buffer, "%d", Processore->NBeforeEscape);
 WritePrivateProfileString("DEBUGGER", "NBeforeEscape", buffer, IniFile);

 TDialog::CloseWindow(n);
}