/* simuliamo il processore e la RAM */

/* Alcune informazioni (ad esempio, come funziona il BCD) le ho prese
   in rete e sono state scritte da:
      -  John West (john@ucc.gu.uwa.edu.au)
      -  Marko Mkel    (msmakela@kruuna.helsinki.fi)
   che devo ricordarmi di ringraziare.
*/


/* da controllare:
   - l'ordine dei flag e' probabilmente sbagliato
   - non sono sicuro di come cresce lo stack
   - Controllare le operazioni aritmetiche in modo decimale.
   - decidere cosa fare in caso di BRK
   - scoprire qual'e' l'ordine esatto (H-L oppure L-H) delle PUSH e delle
     POP per le istruzioni JSR e RTS;
   - far funzionare correttamente gli interrupt e la RTI;
   - simulare le operazioni non documentate.

*/

#include <stdlib.h>
#include <windows.h>
#include "opcodes.h"
#include "process.h"

#define STACK_OFFSET 0x0100
#define TRUE 1
#define FALSE 0
#define HI_BYTE(x)  (unsigned char) ((x & 0xFF00) >> 8)
#define LO_BYTE(x)  (unsigned char) (x & 0x00FF)
#define DEFAULT_RESTART_ADDRESS 0x1000

unsigned char huge RAM[65536]; // memoria RAM
extern unsigned int org_value; // ORG del primo file di un progetto
extern char IniFile[];         // file .INI
extern "C" char decodifica(unsigned char , char *, char *);
extern HWND FinestraPrincipale;

const char ini_sect_debug[] = "DEBUGGER";

T6502::T6502()
{
 //Leggo il file .INI
 OrgFromCode = GetPrivateProfileInt(ini_sect_debug,
		"OrgFromCode", 0, IniFile);
 DefaultRestartAddress = GetPrivateProfileInt(ini_sect_debug,
		"DefaultRestartAddress", DEFAULT_RESTART_ADDRESS, IniFile);
 ResetToOrg = GetPrivateProfileInt(ini_sect_debug,
		"ResetToOrg", 0, IniFile);
 StopAfterReset = GetPrivateProfileInt("DEBUGGER",
		"StopAfterReset", 1, IniFile);
 StopOnBreak = GetPrivateProfileInt("DEBUGGER",
		"StopOnBreak", 1, IniFile);
 ClearRegsOnReset = GetPrivateProfileInt("DEBUGGER",
		"ClearRegsOnReset", 0, IniFile);
 ClearStackOnReset = GetPrivateProfileInt("DEBUGGER",
		"ClearStackOnReset", 0, IniFile);
 ClearFlagsOnReset = GetPrivateProfileInt("DEBUGGER",
		"ClearFlagsOnReset", 1, IniFile);
 CheckStackOverflow = GetPrivateProfileInt("DEBUGGER",
		"CheckStackOverflow", 1, IniFile);
 NBeforeEscape = GetPrivateProfileInt("DEBUGGER",
		"NBeforeEscape", 20, IniFile);

 StackOverflow = FALSE;

 reset();
}

void T6502::reset(void) 
{
 if (ResetToOrg)  
    {
     if(OrgFromCode)
        PC = org_value;               // vai a ORG
     else
        PC=DefaultRestartAddress;     // vai a D.R.A.
    }
 else
     PC = RAM[0xFFFC]+0x100*RAM[0xFFFD]; // reset come l'hardware

 if(ClearRegsOnReset)
       A = X = Y = 0;
 if(ClearStackOnReset)
       SP = 0xFF;
 if(ClearFlagsOnReset)
       {
	//Imposto i valori di default per i flags
	SR.byte = 0;
        SR.bit.X = 1;
       }
}

void T6502::NMI(void)
{
 push(LO_BYTE(PC));
 push(HI_BYTE(PC));
 push(SR.byte);
 SR.bit.I = 1;
 PC=RAM[0xFFFA]+0x100*RAM[0xFFFB];
}

void T6502::IRQ(void)
{
 if(!(SR.bit.I))
  {
   push(LO_BYTE(PC));
   push(HI_BYTE(PC));
   push(SR.byte);
   SR.bit.I = 1;
   PC=RAM[0xFFFE]+0x100*RAM[0xFFFF];
  }
}

void T6502::push(unsigned char n)
{
 if( !SP && CheckStackOverflow )
   StackOverflow = TRUE;
 RAM[STACK_OFFSET+SP]=n;
 SP--;
}

unsigned char T6502::pop(void)
{
 SP++;
 if( !SP && CheckStackOverflow)
      StackOverflow = TRUE;
 return((unsigned char)RAM[STACK_OFFSET+SP]);
}

/* esegue un' istruzione. Ritorna FALSE se si puo' proseguire con l'istruzione
   successiva; altrimenti ritorna:
	    1- per il BRK
	    2- per un opcode errato
*/
int T6502::step()
{
 unsigned char opcode;
 char istr,modo;
 unsigned int valore;

 /* acchiappa l'opcode e aggiunge 1 a PC*/
 opcode = RAM[PC++];
 /* cerca l'opcode (0 - 55) ed il modo di indirizzamento (0-12) */

 /* Per rendere piu' veloce l'emulatore riscrivo in assembler:

 for(istr=0; (istr<56 && !trovato); istr++)
    for(modo=0; (modo<13 && !trovato); modo++)
	   trovato = (op_table[istr][modo] == opcode);

 
 asm		push AX;
 asm		push BX;
 asm		push CX;
 asm		mov CL, 13;
 asm   		mov AX, 55;   //parto al contrario
 asm		mov istr, AX;
      loop_i:
 asm		mov AX, 12;
 asm		mov modo, AX;

      loop_m:				
 asm		mov AX, istr;
 asm		imul CL;
 asm		add AX, modo;
 asm		mov BX, AX;
 asm    	mov AX, word ptr [op_table+BX];
 asm		mov BL, opcode;
 asm		sub AX, BX;
 asm		mov trovato, AL;   //si puo' migliorare, ma devo cambiare anche il resto

 // sottinteso: mov AL, trovato;
 asm	        jz endloop;
 asm		dec modo;
 asm		jge  loop_m;
      end_m:
 asm		dec istr;
 asm		jl endloop;
 asm		jmp loop_i;

      endloop:
 asm		pop CX;
 asm		pop BX;
 asm		pop AX;

*/
 if(!decodifica(opcode, &istr, &modo))
       return(2);  /* illegal opcode */
 /* carica i bytes necessari, aggiorna PC e calcola il valore dell'operando */
 else
      {
       istr--;
       modo--;  /* lo so che non e' elegante. Chi se ne frega. */
      }

 switch(modo)
 {
  case 3:       /* accumulatore */
  case 4:       /* implicito */
	 break;
  case 0:       /* immediato */
  case 2:       /* pag. zero */
  case 5:       /* (IND,X) */
  case 6:       /* (IND),Y */
  case 7:       /* pag.zero,X */
  case 8:       /* pag.zero,Y */
  case 12:      /* relativo (per i branch ) */
	 valore = RAM[PC++];
	 break;
  case 1:       /* assoluto */
  case 9:       /* assoluto,X */
  case 10:      /* assoluto,Y */
  case 11:      /* indiretto (per il JMP) */
	valore = RAM[PC] + 0x0100*RAM[PC+1];
	PC+=2;
	break;
 }
 execute(istr, modo, valore);
 if (StackOverflow)
    { 
     StackOverflow = FALSE;
     return(3); //segnala un'interruzione utente
    }
 if(istr==BRK)
     return(1);
 else
     return(0);
}

void T6502::execute(unsigned char istruzione, int modo, unsigned int valore)
{
 unsigned int locazione;

/* a seconda del modo di indirizzamento, calcola la locazione di memoria
   interessata (tranne immediato, accumulatore, implicito e relativo) */
 switch(modo)
 {
  case 0:       /* immediato */
  case 3:       /* accumulatore */
  case 4:       /* implicito */
  case 12:      /* relativo (per i branch) */
	 break;
  case 1:       /* assoluto */
  case 2:       /* pag. zero */
	 locazione = valore; break;
  case 5:       /* (IND,X) */
	 locazione = RAM[valore+X]+0x0100*RAM[valore+X+1]; break;
  case 6:       /* (IND),Y */
	 locazione = RAM[valore]+0x0100*RAM[valore+1]+Y ; break;
  case 7:       /* pag.zero,X */
  case 9:       /* assoluto,X */
	 locazione = valore+X; break;
  case 8:       /* pag.zero,Y */
  case 10:      /* assoluto,Y */
	 locazione = valore+Y; break;
  case 11:      /* indiretto (per il JMP) */
   /*NOTA: questa e' la versione corretta. In realta' sappiamo che esiste
     un bug nelle versioni piu' vecchie del 6502/6510  */
	 locazione = RAM[valore]+0x0100*RAM[valore+1]; break;
  default:
         MessageBox(NULL, "Internal error",NULL, MB_OK);
 }
       /* ad ogni istruzione la sua @#$%! operazione... */
 switch(istruzione)
 {
  unsigned int temp;
  /*unsigned int op16; */
  unsigned char op8;

  case ADC :
            /* Difficile. Bisogna gestire anche il BCD!! */
	    if(modo==0)
                op8=valore;
	    else
		op8=RAM[locazione];
	    if(SR.bit.D)       // somma in BCD
	       {
	        unsigned
	           AL, /* low nybble of accumulator */
                   AH; /* high nybble of accumulator */

                AL = (A & 15) + (op8 & 15) + SR.bit.C;         /* Calculate the lower nybble. */
                AH = (A >> 4) + (op8 >> 4) + (AL > 15); /* Calculate the upper nybble. */
		if (AL > 9)
                  {
		   AL += 6;                  /* BCD fixup for lower nybble. */
		   AH ++;  /* Questo l'ho aggiunto io. Sembra giusto... */
                  }
		SR.bit.Z = (((A + op8 + SR.bit.C) & 255) != 0);         /* Zero flag is set just
                                                         like in Binary mode. */

             /* Negative and Overflow flags are set with the same logic than in
                Binary mode, but after fixing the lower nybble. */

                SR.bit.N = ((AH & 8) != 0);
                SR.bit.V = ((AH << 4) ^ A) & 128 && !((A ^ op8) & 128);

		if (AH > 9) AH += 6;                  /* BCD fixup for upper nybble. */


             /* Carry is the only flag set after fixing the result. */

                SR.bit.C = (AH > 15);
                A = ((AH << 4) | (AL & 15));

	       }
	    else               // somma in binario
               {
                temp = A+op8;
	        if(SR.bit.C) temp++;
		SR.bit.C = (temp>0xFF);
		SR.bit.V = ( ((signed char)A +(signed char)op8) == (signed char)temp);
		A = temp & 0xFF;
		SR.bit.N = (A & 0x80);
		SR.bit.Z = !A;
	       }
	    break;

  case AND :
	    if(modo==0)
		A&=valore;
	    else
		A&=RAM[locazione];
	    SR.bit.N = (A & 0x80);      /* da cambiare!!!  */
	    SR.bit.Z = !A;
	    break;
  case ASL :
	    if(modo==3)
		{
		 temp=A<<1;
		 A=temp & 0xFF;
		}
	    else
		{
		 temp=RAM[locazione]<<1;
		 RAM[locazione]=temp &0xFF;
		}
	    SR.bit.C = temp & 0x0100;
	    SR.bit.N = temp & 0x80;    
	    SR.bit.Z = !(temp & 0xFF);
	    break;

  case BCC :
	    if(!SR.bit.C)
		PC+= (signed char) valore;
	    break;

  case BCS :
	    if(SR.bit.C)
		PC+= (signed char) valore;
	    break;

  case BEQ :
	    if(SR.bit.Z)
		PC+= (signed char) valore;
	    break;

  case BIT :
	    temp = RAM[locazione] & A;
	    SR.bit.Z = !temp;
	    SR.bit.N = RAM[locazione] & 0x80;
	    SR.bit.V = RAM[locazione] & 0x40;
	    break;

  case BMI :
	    if(SR.bit.N)
		PC+= (signed char) valore;
	    break;

  case BNE :
	    if(!SR.bit.Z)
		PC+= (signed char) valore;
	    break;

  case BPL :
	    if(!SR.bit.N)
		PC+= (signed char) valore;
	    break;

  case BRK :
	    if(StopOnBreak)
		 {    // si comporta some ad un breakpoint
		 }
	    else
		 {    //  si comporta com il 6502 hardware
		      // ( push(PC), push(SR) eccetera... )
		  push(LO_BYTE(PC));
		  push(HI_BYTE(PC));
                  push(SR.byte);
		  PC=RAM[0xFFFE]+0x100*RAM[0xFFFF];
		  SR.bit.I = 1;
                  SR.bit.B = 1;
		 }
	    break;

  case BVC :
	    if(!SR.bit.V)
		PC+= (signed char) valore;
	    break;

  case BVS :
	    if(SR.bit.V)
		PC+= (signed char) valore;
	    break;

  case CLC :
	    SR.bit.C = FALSE;
	    break;

  case CLD :
	    SR.bit.D = FALSE;
	    break;

  case CLI :
	    SR.bit.I = FALSE;
	    break;

  case CLV :
	    SR.bit.V = FALSE;
	    break;

  case CMP :    // Usiamo lo stesso codice di SBC.
	    if(modo==0)
		op8=valore;
	    else
		op8=RAM[locazione];

	    temp = A-op8-(!SR.bit.C);
	    SR.bit.C = (temp & 256) != 0;
	    SR.bit.Z = (temp & 255) == 0;
	    SR.bit.V = (temp ^ op8) & 128 && (A ^ op8) & 128;
	    SR.bit.N = (temp & 128) != 0;

	    break;     

  case CPX :
	    if(modo==0)
		op8=valore;
	    else
		op8=RAM[locazione];

	    temp = X-op8-(!SR.bit.C);
	    SR.bit.C = (temp & 256) != 0;
	    SR.bit.Z = (temp & 255) == 0;
	    SR.bit.V = (temp ^ op8) & 128 && (X ^ op8) & 128;
	    SR.bit.N = (temp & 128) != 0;

	    break;     

  case CPY :
	    if(modo==0)
		op8=valore;
	    else
		op8=RAM[locazione];

	    temp = X-op8-(!SR.bit.C);
	    SR.bit.C = (temp & 256) != 0;
	    SR.bit.Z = (temp & 255) == 0;
	    SR.bit.V = (temp ^ op8) & 128 && (X ^ op8) & 128;
	    SR.bit.N = (temp & 128) != 0;

	    break;     

  case DEC :
	    if (SR.bit.D)
	      {
	       unsigned char Hi, Lo;

	       Lo = (RAM[locazione] & 0x0F)-1;
               Hi = (RAM[locazione] & 0xF0);
	       if (Lo >0x09)
		{
		 Lo = 0x09;
		 Hi -=0x10;
		 if (Hi>0x90) Hi = 0x90;		        
		}
	       RAM[locazione]= temp = Lo&Hi & 0xFF;
	      }
	    else
	       temp = (--RAM[locazione]) & 0xFF;
	    SR.bit.Z=(!temp);
	    SR.bit.N = temp & 0x80;
	    break;

  case DEX :
	    if (SR.bit.D)
	      {
	       unsigned char Hi, Lo;

	       Lo = (X & 0x0F)-1;
               Hi = (X & 0xF0);
	       if (Lo >0x09)
		{
		 Lo = 0x09;
		 Hi -=0x10;
		 if (Hi>0x90) Hi = 0x90;		        
		}
	       X = Lo&Hi & 0xFF;
	      }
	    else
	       X--;

	    SR.bit.Z=(!X);
	    SR.bit.N = X & 0x80;
	    break;

  case DEY :
  	    if (SR.bit.D)
	      {
	       unsigned char Hi, Lo;

	       Lo = (Y & 0x0F)-1;
               Hi = (Y & 0xF0);
	       if (Lo >0x09)
		{
		 Lo = 0x09;
		 Hi -=0x10;
		 if (Hi>0x90) Hi = 0x90;		        
		}
	       Y = Lo&Hi & 0xFF;
	      }
	    else
	       Y--;

	    SR.bit.Z=(!Y);
	    SR.bit.N = Y & 0x80;
	    break;

  case EOR :
	    if(modo==0)
		 temp = valore;
	    else
		 temp = RAM[locazione];
	    A^=temp;
	    SR.bit.N = A & 0x80;
	    SR.bit.Z = !A;
	    break;

  case INC :
  	    if (SR.bit.D)
	      {
	       unsigned char Hi, Lo;

	       Lo = (RAM[locazione] & 0x0F)+1;
               Hi = (RAM[locazione] & 0xF0);
	       if (Lo >0x09)
		{
		 Lo = 0x00;
		 Hi +=0x10;
		 if (Hi>0x90) Hi = 0x00;		        
		}
	       RAM[locazione]= temp = Lo&Hi & 0xFF;
	      }
	    else
	      temp = (++RAM[locazione]) & 0xFF;

	    SR.bit.N = A & 0x80;
	    SR.bit.Z = !temp;
	    break;

  case INX :
	    if (SR.bit.D)
	      {
	       unsigned char Hi, Lo;

	       Lo = (X & 0x0F)+1;
               Hi = (X & 0xF0);
	       if (Lo >0x09)
		{
		 Lo = 0x00;
		 Hi +=0x10;
		 if (Hi>0x90) Hi = 0x00;		        
		}
	       X = Lo&Hi & 0xFF;
	      }
	    else
	       X++;

	    SR.bit.N = X>>7;
	    SR.bit.Z = !X;
	    break;

  case INY :
  	    if (SR.bit.D)
	      {
	       unsigned char Hi, Lo;

	       Lo = (Y & 0x0F)+1;
               Hi = (Y & 0xF0);
	       if (Lo >0x09)
		{
		 Lo = 0x00;
		 Hi +=0x10;
		 if (Hi>0x90) Hi = 0x00;		        
		}
	       Y = Lo&Hi & 0xFF;
	      }
	    else
	       Y++;

	    SR.bit.N = Y>>7;
	    SR.bit.Z = !Y;
	    break;

  case JMP :
	    PC=locazione;
	    break;

  case JSR :
	    push(LO_BYTE(PC));
	    push(HI_BYTE(PC));
	    PC=locazione;
	    break;

  case LDA :
	    if(modo==0)
		   A=valore;
	    else
		   A=RAM[locazione];
	    SR.bit.N = A & 0x80;
	    SR.bit.Z = !A;
	    break;

  case LDX :
	    if(modo==0)
		   X=valore;
	    else
		   X=RAM[locazione];
	    SR.bit.N = X & 0x80;
	    SR.bit.Z = !X;
	    break;

  case LDY :
	    if(modo==0)
		   Y=valore;
	    else
		   Y=RAM[locazione];
  	    SR.bit.N = Y & 0x80;
	    SR.bit.Z = !Y;
	    break;

  case LSR :if(modo==3)
		{
		 SR.bit.C = (A & 0x01);
		 temp = (A>>=1);
		}
	    else
		{
		 SR.bit.C = (RAM[locazione] & 0x01);
		 temp = (RAM[locazione] >>=1);
		}
	    SR.bit.N = FALSE;
	    SR.bit.Z = !temp;
	    break;

  case NOP :
	    break;

  case ORA :
	    if(modo==0)
		A|=valore;
	    else
		A|=RAM[locazione];
	    SR.bit.N = A & 0x80;
	    SR.bit.Z = !A;
	    break;

  case PHA :
	    push(A);
	    break;

  case PHP :
	    push(SR.byte);
	    break;

  case PLA :
	    A=pop();
	    SR.bit.N = A & 0x80;
	    SR.bit.Z = !A;
	    break;

  case PLP :
	    SR.byte=pop();
	    break;

  case ROL :
	    if(modo == 3)
		{
		 SR.bit.C= A & 0x80;
		 A<<=1;
		 temp = A|=SR.bit.C;
		}
	    else
		{
		 SR.bit.C= RAM[locazione] & 0x80;
		 RAM[locazione]<<=1;
		 temp = RAM[locazione] |= SR.bit.C;
		}
	    SR.bit.N = temp & 0x80;
	    SR.bit.Z = !temp;
	    break;

  case ROR :
	    if(modo == 3)
		{
		 SR.bit.C= A & 0x01;
		 A>>=1;
		 A|=(SR.bit.C*0x80);
		}
	    else
		{
		 SR.bit.C= RAM[locazione] & 0x01;
		 RAM[locazione]>>=1;
		 RAM[locazione] |= (SR.bit.C*0x80);
		}
	    SR.bit.N = temp & 0x80;
	    SR.bit.Z = !temp;
	    break;

  case RTI :
             SR.byte = pop();
	     PC=0x0100*pop() + pop();
             SR.bit.I = 0;
	     break;

  case RTS :
	     PC=0x0100*pop() + pop();
	     break;

  case SBC :
	    /* Difficile. Bisogna gestire anche il BCD!! */
	    if(modo==0)
                op8=valore;
	    else
		op8=RAM[locazione];
	    if(SR.bit.D)       // sottrazione BCD
	       {
		unsigned AL,AH;

                AL = (A & 15) - (op8 & 15) - !SR.bit.C;        /* Calculate the lower nybble. */
                if (AL & 16) AL -= 6;                 /* BCD fixup for lower nybble. */
                AH = (A >> 4) - (op8 >> 4) - (AL & 16); /* Calculate the upper nybble. */
                if (AH & 16) AH -= 6;                 /* BCD fixup for upper nybble. */
                       /* The flags are set just like in Binary mode. */
		temp = A - op8 - !SR.bit.C;
		SR.bit.C = (temp & 256) != 0;
		SR.bit.Z = (temp & 255) != 0;
		SR.bit.V = (temp ^ op8) & 128 && (A ^ op8) & 128;
                SR.bit.N = (temp & 128) != 0;
                A = ((AH << 4) | (AL & 15)) & 255;
	       }
	    else               // somma in binario
               {
                A = A-op8;
	        if(!SR.bit.C) A--;
		SR.bit.C = (temp>0xFF);
		SR.bit.V = ( ((signed char)A -(signed char)op8) == (signed char)temp);
		A = temp & 0xFF;
		SR.bit.N = (A & 0x80);
		SR.bit.Z = !A;
	       }
	    break;

  case SEC : SR.bit.C=TRUE;
	     break;

  case SED : SR.bit.D=TRUE;
	     break;

  case SEI :
	     SR.bit.I=TRUE;
	     break;

  case STA :
	     RAM[locazione] = A;
	     break;

  case STX :
	     RAM[locazione] = X;
	     break;

  case STY :
	     RAM[locazione] = Y;
	     break;

  case TAX :
	     X=A;
	     SR.bit.N = X & 0x80;
	     SR.bit.Z = !X;
	     break;

  case TAY :
	     Y=A;
	     SR.bit.N = Y & 0x80;
	     SR.bit.Z = !Y;
             break;

  case TSX :
	     X=SP;
   	     SR.bit.N = X & 0x80;
	     SR.bit.Z = !X;
	     break;

  case TXA :
	     A=X;
	     SR.bit.N = A & 0x80;
	     SR.bit.Z = !A;
	     break;

  case TXS :
	     SP=X;
	     SR.bit.N = SP & 0x80;
	     SR.bit.Z = !SP;
	     break;

  case TYA :
	     A=Y;
	     SR.bit.N = A & 0x80;
	     SR.bit.Z = !A;
	     break;

  default:
	   exit(1);
 }
}