/* Classe TProjectWindow */

#include <filewnd.h>
#include <string.h>
#include <stdio.h>
#include <list.h>
#include <commdlg.h>
#include "source.h"
#include "project.h"
#include "usermsg.h"

//dichiara la funzione (esterna) dell'assemblatore...
extern "C" void winasm(char *, char *);
//funzione per liberare la memoria della watch table
extern "C" void delete_all_ws(); 
//Dichiara l'handle della finestra padre della ListBox a cui mandare i messaggi:
extern HWND FinestraMessaggi;
extern HWND FinestraPrompt;
extern char org_link_flag;

const char ProjectFilter[] = "Project file (*.pro)\0 *.pro\0";

TProjectList::TProjectList(PTWindowsObject AParent, int AnId, int X, int Y,
		           int W, int H, PTModule AModule = NULL)
	     : TListBox(AParent, AnId, X, Y, W, H, AModule)
{
  Attr.Style = WS_CHILD | WS_VISIBLE | WS_GROUP | WS_TABSTOP | WS_VSCROLL
	      | LBS_NOTIFY | LBS_NOINTEGRALHEIGHT;
}

int TProjectList::AddString(char *nome)
{
 return TListBox::AddString(nome);
}


TProjectWindow::TProjectWindow(PTWindowsObject AParent, TProjectWindow **Me, TMessageWindow *msgwnd)
      : TWindow(AParent, "Project")
{
 HWND handle;
 HMENU hm;

 Myself = Me;
 Messages = msgwnd;
 Elenco = new TProjectList(this, ID_PROJECTLIST, 1,1,1,1);
 FileName[0] = '\0';
 first_project_file = NULL;

 //Abilita la voce del menu.
 handle = GetApplication()->MainWindow->HWindow;
 hm = GetMenu(handle);
 EnableMenuItem(hm, CM_PROGETTO, MF_BYCOMMAND | MF_ENABLED);
 DrawMenuBar(handle);
}

TProjectWindow::~TProjectWindow()
{
 project_file *pt;

 // cancella la lista dei files del progetto

 while(first_project_file)
    {
     pt= first_project_file;
     first_project_file = first_project_file->next;
     free(pt->name) ;
     free(pt->path);
     free(pt);
    }

}

void TProjectWindow::SetupWindow()
{
 HFONT f;

 TWindow::SetupWindow();
 f = CreateFont(-12, 0, 0, 0, FW_BOLD, 0, 0, 0, 0, 0, 0, 0, 0, "MS Sans Serif");
 SendMessage(Elenco->HWindow, WM_SETFONT, (WPARAM)f, (LPARAM) MAKELONG((WORD) TRUE, 0) );
 SetName();
 Refresh();
}

void TProjectWindow::WMSize(RTMessage msg)
{
 TWindow::WMSize(msg);
 if (msg.WParam != SIZE_MINIMIZED)
      MoveWindow(Elenco->HWindow, 0, 0, msg.LP.Lo, msg.LP.Hi, TRUE);
}

LPSTR TProjectWindow::GetClassName()
{
 return("ProjectWindow");
}

// Assegna l'icona della finestra MDI
void TProjectWindow::GetWindowClass(WNDCLASS& AWndClass)
{
  TWindow::GetWindowClass(AWndClass);
  AWndClass.hIcon = LoadIcon(GetApplication()->hInstance, "ICONA_PROGETTO");
}

// Non si distrugge ma si nasconde solo.
void TProjectWindow::WMClose(RTMessage )
{
 Destroy(); // distrugge l'oggetto di interfaccia, ma non
            // la lista dei files 

 //Show(SW_HIDE);
}

// Aggiungi un nome di file dopo quello selezionato
int TProjectWindow::AddString(LPSTR s)
{
 project_file *pt;
 project_file *ptprev;
 char name[MAXFILE];
 char ext[MAXEXT];
 int n,i;

 pt = (project_file *)malloc(sizeof(project_file));
 pt->path = (char *)malloc(strlen(s)+1);
 pt->name = (char *)malloc(MAXFILE+MAXEXT+1);

 strcpy(pt->path, s);
 fnsplit(s, NULL, NULL, name, ext);
 strncpy(pt->name, name, MAXFILE);
 strcat(pt->name, ext);  

 n=Elenco->GetSelIndex();
 ptprev = first_project_file;

 if ( (!ptprev) || (n==0) )  // lista vuota o primo elemento della lista
       {
	pt->next = first_project_file;
        first_project_file = pt;
       }
 else
       {
	if (n<0)  //nessun elemento selezionato -> aggiungo in coda
	     while(ptprev->next)
		  ptprev = ptprev->next;

        else      //selezionato elemento (n-1)-> inserisco
	     {
              i=1;
	      while( ptprev->next && (i<n) )
                 {
                  ptprev = ptprev->next;
                  i++;
	         }
	     }
	pt->next = ptprev->next;
        ptprev->next = pt;
       }

 return( TRUE );
}

// Aggiungi un nome di file alla posizione n
// (se n<0 aggiunge in coda)

int TProjectWindow::AddString(LPSTR s, int n)
{
 project_file *pt;
 project_file *ptprev;
 char name[MAXFILE];
 char ext[MAXEXT];
 int i;

 pt = (project_file *)malloc(sizeof(project_file));
 pt->path = (char *)malloc(strlen(s)+1);
 pt->name = (char *)malloc(MAXFILE+MAXEXT+1);

 strcpy(pt->path, s);
 fnsplit(s, NULL, NULL, name, ext);
 strncpy(pt->name, name, MAXFILE);
 strcat(pt->name, ext);  

 ptprev = first_project_file;

 if ( (!ptprev) || (n==0) )  // lista vuota o primo elemento della lista
       {
	pt->next = first_project_file;
        first_project_file = pt;
       }
 else
       {
	if (n<0)  //nessun elemento selezionato -> aggiungo in coda
	     while(ptprev->next)
		  ptprev = ptprev->next;

        else      //selezionato elemento (n-1)-> inserisco
	     {
              i=1;
	      while( ptprev->next && (i<n) )
                 {
                  ptprev = ptprev->next;
                  i++;
	         }
	     }
	pt->next = ptprev->next;
        ptprev->next = pt;
       }

 return( TRUE );
}


// Cancella il file selezionato(se uno e' selezionato).
int TProjectWindow::DeleteSelected()
{
 project_file *pt = first_project_file;
 project_file *pt2;
 int n,i;

 n=Elenco->GetSelIndex();
 if(n<0)
     return(-1);
 else
     {
      if (n==0)
         {
	  free(pt->name);
	  free(pt->path);
          first_project_file = pt->next;
	 } 
      else
         {
          for(i=1; i<n; i++)
	      pt=pt->next;
	  free(pt->next->name);
	  free(pt->next->path);
          pt2 = pt->next->next;
          free(pt->next);
	  pt->next=pt2;
	 }
     }
 return (Elenco->DeleteString(n));
}

// Apre un progetto esistente
void TProjectWindow::OpenProject()
{
  static OPENFILENAME ofn;
  char szFile[256], szFileTitle[256];

  szFile[0] = '\0';

  /* Set all structure members to zero. */

  memset(&ofn, 0, sizeof(OPENFILENAME));

  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = HWindow;
  ofn.lpstrFilter = ProjectFilter;
  ofn.nFilterIndex = 1;
  ofn.lpstrFile= szFile;
  ofn.nMaxFile = sizeof(szFile);
  ofn.lpstrFileTitle = szFileTitle;
  ofn.nMaxFileTitle = sizeof(szFileTitle);
  ofn.lpstrInitialDir = NULL;
  ofn.Flags = OFN_HIDEREADONLY| OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

  if (GetOpenFileName(&ofn))
     {
      strcpy(FileName, ofn.lpstrFile);
      Read();
     }
}

// Salva e chiude il progetto.
BOOL TProjectWindow::CloseProject()
{
 HWND handle;
 HMENU hm;
 project_file *pt = first_project_file;

 if(!Save())
   return (TRUE);
  
 // cancella la lista dei files del progetto

 while(first_project_file)
    {
     pt= first_project_file;
     first_project_file = first_project_file->next;
     free(pt->name) ;
     free(pt->path);
     free(pt);
    }

 *Myself = NULL;

 //Disabilita la voce del menu.
 handle = GetApplication()->MainWindow->HWindow;
 hm = GetMenu(handle);
 EnableMenuItem(hm, CM_PROGETTO, MF_BYCOMMAND | MF_GRAYED);
 DrawMenuBar(handle);

 Destroy();
 delete this;
 return(FALSE);
}

// La finestra di progetto puo' assemblare autonomamente
// il file selezionato
void TProjectWindow::Assemble(RTMessage)
{
 PTDialog pt;
 char path[MAXPATH], drive[MAXDRIVE], dir[MAXDIR];
 char name[MAXFILE], ext[MAXEXT];
 char path_out[MAXPATH];
 int n,i;
 project_file *nextpt;

 //if ( Elenco->GetSelString(path, MAXPATH)>1 )
 if ( (n=Elenco->GetSelIndex())>=0 )
   {
    nextpt = first_project_file;
    for(i=0; i<n; i++)
       if(nextpt)
          nextpt=nextpt->next;
    if (!nextpt)
	  return;
    strcpy(path, nextpt->path);

    if(!Messages->HWindow)
       GetModule()->MakeWindow(Messages);
    Messages->Show(SW_SHOW);

    pt=new TDialog(this, ID_PROMPTWINDOW);
    GetApplication()->MakeWindow(pt);

    Messages->Cancella();
    FinestraMessaggi = Messages->HWindow;
    FinestraPrompt = pt->HWindow; 

    fnsplit(path, drive, dir, name, ext);
    strcpy(ext, "OUT");
    fnmerge(path_out, drive, dir, name, ext);

    delete_all_ws();  //cancella la tabella dei simboli di watch
    org_link_flag = FALSE; // un file solo -> occorre ORG
    winasm(path, path_out);

    pt->CloseWindow();
    //Rinfresca le finestre (RAM, stack, disassembler...)
    SendMessage(Parent->HWindow, WM_USER + REFRESH_ALL, 0, 0L); 
   }
}

// Build all. Assembla tutti i files del progetto, tenendo conto
// (quando avremo deciso come... ) dei nomi esterni.
void TProjectWindow::BuildAll(RTMessage)
{
 PTDialog pt;
 char path[MAXPATH], drive[MAXDRIVE], dir[MAXDIR];
 char name[MAXFILE], ext[MAXEXT];
 char path_out[MAXPATH];
// int i, n;
 project_file *nextfile;

 //if( (n=Elenco->GetCount()) > 0)
 if(first_project_file) //se c'e' almeno un file nel progetto...
    {
     if(!Messages->HWindow)
          GetModule()->MakeWindow(Messages);
     Messages->Show(SW_SHOW);

     pt=new TDialog(this, ID_PROMPTWINDOW);
     GetApplication()->MakeWindow(pt);

     Messages->Cancella();
     FinestraMessaggi = Messages->HWindow;
     FinestraPrompt = pt->HWindow;

     delete_all_ws();  //cancella la tabella dei simboli di watch

     nextfile = first_project_file;
     org_link_flag = FALSE; // FALSE per il primo file, TRUE per i successivi
     while(nextfile)
         {
	  strcpy(path, nextfile->path);
	  fnsplit(path, drive, dir, name, ext);
          strcpy(ext, "OUT");
	  fnmerge(path_out, drive, dir, name, ext);
	  winasm(path, path_out);
	  nextfile = nextfile->next;
	  org_link_flag = TRUE; // TRUE per i files successivi al primo
         }
     pt->CloseWindow();
     //Rinfresca le finestre (RAM, stack, disassembler...)
     SendMessage(Parent->HWindow, WM_USER + REFRESH_ALL, 0, 0L);
    }
}

// Modifica il menu della finestra MDI
void TProjectWindow::ActivationResponse(WORD Activated, BOOL IsIconified)
{
 HWND handle;
 HMENU hm;

 TWindow::ActivationResponse(Activated, IsIconified);
 handle = GetApplication()->MainWindow->HWindow;
 hm = GetMenu(handle);
 if(Activated)
     {
//menu File
      EnableMenuItem(hm, CM_FILESAVE,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_FILESAVEAS,  MF_BYCOMMAND | MF_GRAYED);
//menu Edit
      EnableMenuItem(hm, CM_EDITUNDO,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCUT,     MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCOPY,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITPASTE,   MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITDELETE,  MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCLEAR,   MF_BYCOMMAND | MF_GRAYED);
//menu Search
      EnableMenuItem(hm, CM_EDITFIND,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITREPLACE, MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITFINDNEXT,MF_BYCOMMAND | MF_GRAYED);
//menu Assembler
      EnableMenuItem(hm, MENU_ASSEMBLER_ASSEMBLE,MF_BYCOMMAND | MF_GRAYED);
     }
 else
     {
//menu File
      EnableMenuItem(hm, CM_FILESAVE,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_FILESAVEAS,  MF_BYCOMMAND | MF_ENABLED);
//menu Edit
      EnableMenuItem(hm, CM_EDITUNDO,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCUT,     MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCOPY,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITPASTE,   MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITDELETE,  MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCLEAR,   MF_BYCOMMAND | MF_ENABLED);
//menu Search
      EnableMenuItem(hm, CM_EDITFIND,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITREPLACE, MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITFINDNEXT,MF_BYCOMMAND | MF_ENABLED);
//menu Assembler
      EnableMenuItem(hm, MENU_ASSEMBLER_ASSEMBLE,MF_BYCOMMAND | MF_ENABLED);
//Menu Project
      EnableMenuItem(hm, MENU_PROJECT_CLOSE, MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, MENU_PROJECT_ADD, MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, MENU_PROJECT_REMOVE, MF_BYCOMMAND | MF_ENABLED);


     }
 DrawMenuBar(handle);
}

BOOL TProjectWindow::Save()
{
 int WantToSave;

 if (!strlen(FileName))
    {
     WantToSave = MessageBox(HWindow, "Do you want to save your project?", "Closing project...", 0x24 );
     if (WantToSave==6)
	   return(SaveAs());
     else
	   return TRUE;
    }
 else
     return(Write());
}

BOOL TProjectWindow::SaveAs()
{
  static OPENFILENAME ofn;
  char TmpName[MAXPATH];
  char OldName[MAXPATH];
  char szFileTitle[256];

  _fstrcpy(OldName, FileName);
  if ( FileName[0] )
    _fstrcpy(TmpName, FileName);
  else
    TmpName[0] = '\0';

  /* Set all structure members to zero. */
  memset(&ofn, 0, sizeof(OPENFILENAME));

  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = Parent->HWindow;  
  ofn.lpstrFilter = ProjectFilter;
  ofn.lpstrFile= TmpName;
  ofn.nMaxFile = sizeof(TmpName);
  ofn.lpstrFileTitle = szFileTitle;
  ofn.nMaxFileTitle = sizeof(szFileTitle);
  ofn.lpstrInitialDir = NULL;

  ofn.Flags = OFN_SHOWHELP | OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY;

  if (GetSaveFileName(&ofn))
    {
     strcpy(FileName, TmpName);
     if ( Write() )
       return TRUE;
     else
      {
       strcpy(FileName, OldName);
       return FALSE;
      }
    }
  else
    {
     strcpy(FileName, OldName);
     return FALSE;
    }
}

BOOL TProjectWindow::Write()
{
 FILE *f;
 project_file *ps = first_project_file;

 f=fopen(FileName, "w");
 if(!f) return (FALSE);

 while(ps)
    {
     fprintf(f, "%s\n", ps->path);
     ps=ps->next;
    }
 fclose(f);
 return(TRUE);
}

BOOL TProjectWindow::Read()
{
  FILE *f;
  char buffer[256];

  f=fopen(FileName, "r");
  if(!f) return(FALSE);
  while (fgets(buffer, 255, f))
     {
      // elimina il newline
      buffer[strlen(buffer)-1]= '\0';
// non serve per caso...      Elenco->SetSelIndex(Elenco->GetCount()+2);
      AddString(buffer,-1);
     }
  fclose(f);
  Refresh();
  return(TRUE);
}

void TProjectWindow::SetName()
{
 char buffer[255];
 char name [MAXFILE];
 char ext [MAXEXT];

 if(FileName[0])
      fnsplit(FileName, NULL, NULL, name, ext);
 else
     {
      strcpy(name, "untitled");
      strcpy(ext, ".pro");
     }
 sprintf(buffer, "Project: %s%s", name, ext);
 SetCaption(buffer);
}

void TProjectWindow::Refresh()
{
 project_file *pt = first_project_file;

 Elenco->ClearList();
 while(pt)
   {
    Elenco->AddString(pt->name); 
    pt=pt->next;
   }
 Elenco->AddString(""); //per poter inserire un file dopo la fine lista.
}

//Aggiunge un file al progetto
void TProjectWindow::MenuProjectAdd(RTMessage)
{
  static OPENFILENAME ofn;
  char szFile[256], szFileTitle[256];
  const char szFilter[] =  "Source files (*.IN)\0*.in\0Object files (*.OUT)\0*.out\0";
  szFile[0] = '\0';
  char tempstring[256];
  int namepointer, flen, totlen;

  /* Set all structure members to zero. */

  memset(&ofn, 0, sizeof(OPENFILENAME));

  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = HWindow;
  ofn.lpstrFilter = szFilter;
  ofn.nFilterIndex = 1;
  ofn.lpstrFile= szFile;
  ofn.nMaxFile = sizeof(szFile);
  ofn.lpstrFileTitle = szFileTitle;
  ofn.lpstrTitle = "Add files to project";
  ofn.nMaxFileTitle = sizeof(szFileTitle);
  ofn.lpstrInitialDir = NULL;
  ofn.Flags = OFN_ALLOWMULTISELECT | OFN_HIDEREADONLY| OFN_PATHMUSTEXIST
	     | OFN_FILEMUSTEXIST;

// Attenzione signori: possiamo aggiungere piu' di un file alla volta!!!!
//      --> devo scriver un parser per divider i files da aggiunger

  if (GetOpenFileName(&ofn))
     {
      strcpy(tempstring, ofn.lpstrFile);
      tempstring[ofn.nFileOffset-1] = '\\';

      namepointer = ofn.nFileOffset;
      totlen = strlen(ofn.lpstrFile);

      while(namepointer < totlen)
	  {
	   flen = 0;
	   while(ofn.lpstrFile[namepointer+flen]!=' ')
		 {
		  if ( (namepointer + flen) > totlen)
		     break;
		  else
                     flen++;
		 }
	   strncpy(tempstring+ofn.nFileOffset, ofn.lpstrFile+namepointer, flen);
           tempstring[ofn.nFileOffset+flen]='\0';
	   AddString(strlwr(tempstring));
	   namepointer+=flen+1;
	  }       
     }
 // else
 //     ErrorHandler();

 Refresh();
}

//Elimina un file dal progetto
void TProjectWindow::MenuProjectRemove(RTMessage)
{
  DeleteSelected();
}

BOOL TProjectWindow::CanClose()
{
 return(Save() && TWindow::CanClose());
}

// apre l'n-esimo file del progetto, o ne porta in primo piano la finestra

void TProjectWindow::OpenSourceFile(int n)
{
 int i;
 project_file *pt = first_project_file;

 for(i=0; i<n; i++)
     if(pt) pt=pt->next;
 if(pt->path)  // se non e' un NULL:
     SendMessage(Parent->HWindow, WM_USER + EDIT_BY_NAME, 0, (LPARAM)pt->path );
}

void TProjectWindow::WMCommand(RTMessage msg)
{
 if(msg.WP.Lo)
    // se il messaggio arriva da un controllo:
    if(msg.WParam==ID_PROJECTLIST)
       {
	if (msg.LP.Hi == LBN_DBLCLK)
	   OpenSourceFile(Elenco->GetSelIndex()); //forse e' sbagliato
       }

 TWindow::WMCommand(msg);
}
