/* Classe SourceEditor */

#include <filewnd.h>
#include <string.h>
#include <commdlg.h>
#include "source.h"


//dichiara la funzione (esterna) dell'assemblatore...
extern "C" void winasm(char *, char *);
//funzione per liberare la memoria della watch table
extern "C" void delete_all_ws(); 
//Dichiara l'handle della finestra padre della ListBox a cui mandare i messaggi:
extern HWND FinestraMessaggi;
extern HWND FinestraPrompt;

extern char org_link_flag;


TSourceEditor::TSourceEditor(PTWindowsObject AParent, LPSTR ATitle, TMessageWindow *MessageWindow, LPSTR FileName=NULL)
      : TFileWindow(AParent, ATitle, FileName)
{
 Messages = MessageWindow;
}

void TSourceEditor::SetupWindow()
{
 HGDIOBJ f;

 TFileWindow::SetupWindow();
 f=GetStockObject(ANSI_FIXED_FONT);
 SendMessage(Editor->HWindow, WM_SETFONT, (WPARAM)f, (LPARAM) MAKELONG((WORD) TRUE, 0) );
}

LPSTR TSourceEditor::GetClassName()
{
 return("SourceEditor");
}

void TSourceEditor::GetWindowClass(WNDCLASS& AWndClass)
{
  TFileWindow::GetWindowClass(AWndClass);
  AWndClass.hIcon = LoadIcon(GetApplication()->hInstance, "ICONA_SOURCE");
}

BOOL TSourceEditor::CanClear()
{
  char S[MAXPATH+28];
  int Rslt;

  if ( Editor->IsModified() )
  {
    if ( !FileName || (FileName[0]=='\0'))
      _fstrcpy(S, "Untitled file has changed.  Save?");
    else
      wsprintf(S, "File \"%s\" has changed.  Save?", FileName);
    Rslt = MessageBox(HWindow, S, "File Changed",
                          MB_YESNOCANCEL | MB_ICONQUESTION);
    if ( Rslt == IDYES )
      return Save();
    else
      return Rslt != IDCANCEL;
  }
  return TRUE;
}


void TSourceEditor::CMFileSave(RTMessage)
{
 Save();
}

BOOL TSourceEditor::Save()
{
  if ( Editor->IsModified() )
  {
    if ( IsNewFile )
      return SaveAs();
    else
    {
      if ( Write() )
        return TRUE;
      else
        return FALSE;
    }
  }
  else       // Editor's contents haven't been changed. No need to write.
    return TRUE;
}


void TSourceEditor::CMFileSaveAs(RTMessage)
{
 SaveAs();
}

BOOL TSourceEditor::SaveAs()
{
  char TmpName[MAXPATH];
  char OldName[MAXPATH];

  static OPENFILENAME ofn;
  char szFileTitle[256];
  char  szFilter[] = "Source (*.in)\0 *.in\0All files\0*.*\0";

  _fstrcpy(OldName, FileName);
  if ( FileName )
    _fstrcpy(TmpName, FileName);
  else
    TmpName[0] = '\0';

  /* Set all structure members to zero. */
  memset(&ofn, 0, sizeof(OPENFILENAME));

  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = HWindow;
  ofn.lpstrFilter = szFilter;
  ofn.lpstrFile= TmpName;
  ofn.nMaxFile = sizeof(TmpName);
  ofn.lpstrFileTitle = szFileTitle;
  ofn.nMaxFileTitle = sizeof(szFileTitle);
  ofn.lpstrInitialDir = NULL;

  ofn.Flags = OFN_SHOWHELP | OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY;

  if (GetSaveFileName(&ofn))
    {
     SetFileName(TmpName);
     if ( Write() )
       return TRUE;
     else
      {
       SetFileName(OldName);
       return FALSE;
      }
    }
  else
    {
     SetFileName(OldName);
     return FALSE;
    }
}


// L'editor deve essere in grado di assemblare autonomamente
// il proprio file.
void TSourceEditor::Assemble(RTMessage)
{
 PTDialog pt;
 char path[MAXPATH], drive[MAXDRIVE], dir[MAXDIR];
 char name[MAXFILE], ext[MAXEXT];

 if(!Messages->HWindow)
      GetModule()->MakeWindow(Messages);
 Messages->Show(SW_SHOW);

 pt=new TDialog(this, ID_PROMPTWINDOW);
 GetApplication()->MakeWindow(pt);

 Messages->Cancella();
 FinestraMessaggi = Messages->HWindow;
 FinestraPrompt = pt->HWindow; 

 strcpy(path, FileName);
 fnsplit(path, drive, dir, name, ext);
 strcpy(ext, "OUT");
 fnmerge(path, drive, dir, name, ext);

 delete_all_ws();  //cancella la tabella dei simboli di watch
 org_link_flag = FALSE; // assemblo un file solo -> DEVE esserci ORG
  
 winasm(FileName,path);
 //Rinfresca le finestre (RAM, stack, disassembler...)
 SendMessage(Parent->HWindow, WM_USER + REFRESH_ALL, 0, 0L);

 pt->CloseWindow();
}

/* Funzione per la ricerca di un nome di file tra le finestre */
void TSourceEditor::IsThisYourFile(RTMessage msg)
{
 if(strcmp((char *)msg.LParam, FileName) )
    msg.Result = NULL;
 else
    msg.Result = (LRESULT)Editor->HWindow;
}

void TSourceEditor::GotoLine(RTMessage msg)
{
 int n;

 n = Editor->GetLineIndex(msg.WParam);
 Editor->SetSelection(n, n);
}

char *TSourceEditor::nameOf() const
{
 return("TSourceEditor");
}

void TSourceEditor::ActivationResponse(WORD Activated, BOOL IsIconified)
{
 HWND handle;
 HMENU hm;

 TWindow::ActivationResponse(Activated, IsIconified);
 handle = GetApplication()->MainWindow->HWindow;
 hm = GetMenu(handle);
 if(Activated)
     {
//menu File
      EnableMenuItem(hm, CM_FILESAVE,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_FILESAVEAS,  MF_BYCOMMAND | MF_ENABLED);
//menu Edit
      EnableMenuItem(hm, CM_EDITUNDO,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCUT,     MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCOPY,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITPASTE,   MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITDELETE,  MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCLEAR,   MF_BYCOMMAND | MF_ENABLED);
//menu Search
      EnableMenuItem(hm, CM_EDITFIND,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITREPLACE, MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITFINDNEXT,MF_BYCOMMAND | MF_ENABLED);
//menu Assembler
      EnableMenuItem(hm, MENU_ASSEMBLER_ASSEMBLE,MF_BYCOMMAND | MF_ENABLED);
     }
 else
     {
//menu File
      EnableMenuItem(hm, CM_FILESAVE,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_FILESAVEAS,  MF_BYCOMMAND | MF_GRAYED);
//menu Edit
      EnableMenuItem(hm, CM_EDITUNDO,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCUT,     MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCOPY,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITPASTE,   MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITDELETE,  MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCLEAR,   MF_BYCOMMAND | MF_GRAYED);
//menu Search
      EnableMenuItem(hm, CM_EDITFIND,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITREPLACE, MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITFINDNEXT,MF_BYCOMMAND | MF_GRAYED);
//menu Assembler
      EnableMenuItem(hm, MENU_ASSEMBLER_ASSEMBLE,MF_BYCOMMAND | MF_GRAYED);
     }
 DrawMenuBar(handle);
}

