/* Questo file contiene il codice delle tre symbol tables:
	      - per l'assembler (simboli risolti)
	      - per l'assembler (simboli irrisolti)
	      - per il watching delle variabili
*/

/*------------------------------------------------------------------*/
/* Symbol table per l'assemblatore*/

/* ogni singola entry della symbol table e' costituita da
	  - un puntatore a stringa (l'identificatore)
	  - un intero (il suo valore)
	  - un puntatore alla prossima entry */

#include <string.h>
#include <stdlib.h>
#include "\borlandc\sorgenti\6502sdk\multiio.h"

#define TRUE 1
#define FALSE 0

typedef struct t_entry{
			char *id;
			int  val;
			struct t_entry *next;
		       }   entry;

entry *first = NULL;

/* aggiungi un elemento alla symbol table */

void add_symbol(char *newid, int newval)
{
 entry *last;

 if(first==NULL)
  {
    first = malloc(sizeof(entry));
    first->next = NULL;
  }
 else
   {
    last=malloc(sizeof(entry));
    last->next=first;
    first = last;
   }
 first->val=newval;
 first->id = malloc(strlen(newid)+1);
 strcpy(first->id, newid);
}

/* cancella TUTTA la symbol table */
void delete_symbol_table (void)
{
 entry *cur;
 while(first!=NULL)
     {
      cur=first;
      first=first->next;
      free (cur->id);
      free(cur);
     }
}

/* mette in 'value' il valore dell'elemento 'nome' della symbol table.
   Ritorna TRUE se l'elemento esiste, altrimenti FALSE.   */

int get_symbol(char *nome, unsigned int *value)
{
 entry *cur;

 cur = first;
 /* attenzione: strcmp ritorna ZERO se s1==s2 ! */

 while(cur !=NULL)
      {
       if(strcmp(nome,cur->id))
	      cur=cur->next;
       else
	      {
	       *value = cur->val;
	       return(TRUE);
	      }
      }
 return(FALSE);
}

/* stampa la symbol table */

void print_symbol_table(void)
{
 entry *cur;
 cur = first;
 io_symecho("\n");
 while (cur!=NULL)
    {
     io_symecho("%*s : %6u = %04X\n", 20, cur->id, cur->val, cur->val);
     cur = cur->next;
    }
 io_symecho("\n");
}

/* genera la symbol table per il debugger nel file 'file_out' */

void gen_symbol_table(FILE *file_out)
{
 entry *cur;
 cur = first;
 while (cur!=NULL)
    {
     fprintf(file_out,"%s\t%04X\n",  cur->id, cur->val);
     cur = cur->next;
    }
}



/*-----------------------------------------------------------------*/


/* Symbol table dei riferimenti irrisolti*/

/* ogni singola entry della symbol table e' costituita da
	  - un puntatore a stringa (l'identificatore)
	  - un puntatore alla prossima entry */

/* la tabella viene scritta solo al passo 0 usando le parti sinistre
   delle direttive EQU irrisolte                                   */

typedef struct t_ientry{
			char *id;
			int  val;
			struct t_ientry *next;
		       }   ientry;

ientry *ifirst = NULL;

/* aggiungi un elemento alla tabella dei simboli irrisolti*/

void add_symbol_i(char *newid)
{
 ientry *ilast;

 if(ifirst==NULL)
  {
    ifirst = malloc(sizeof(ientry));
    ifirst->next = NULL;
  }
 else
   {
    ilast=malloc(sizeof(ientry));
    ilast->next=ifirst;
    ifirst = ilast;
   }
 ifirst->id = malloc(strlen(newid)+1);
 strcpy(ifirst->id,newid);
}

/* cancella TUTTA la symbol table */
void delete_symbol_table_i (void)
{
 ientry *cur;
 while(ifirst!=NULL)
     {
      cur=ifirst;
      ifirst=ifirst->next;
      free (cur->id);
      free(cur);
     }
}

/*  Ritorna TRUE se l'elemento esiste, altrimenti FALSE.   */

int get_symbol_i(char *nome)
{
 ientry *cur;

 cur = ifirst;
 /* attenzione: strcmp ritorna ZERO se s1==s2 ! */

 while(cur !=NULL)
      {
       if(strcmp(nome,cur->id))
	      cur=cur->next;
       else
	      return(TRUE);
      }
 return(FALSE);
}




/*------------------------------------------------------------------------*/
/* Tabella di watch per le variabili.
   Prossimamente magari uso un B-tree che e' piu' veloce, in modo da avere
   anche un refresh veloce delle finestre.
*/

typedef struct t_watch_item
       {
	char *name;
        unsigned int address;
	struct t_watch_item *next;  
       }   watch_item;

watch_item *first_wi;
watch_item *next_wi;

void add_watch_symbol(char *s, unsigned int val)
{
 watch_item *wi;

 wi = malloc(sizeof(watch_item));
 wi->name = malloc(strlen(s)+1);
 strcpy(wi->name, s);
 wi->address = val;
 wi->next = first_wi;
 first_wi = wi;
 next_wi = wi;   
}

char get_watch_symbol(char *s, unsigned int *val)
{
 watch_item *wi = first_wi;

 while(wi)
    {
     if(!strcmp(wi->name, s)) break;
     wi = wi->next;
    }
 if(!wi)
    return(FALSE);

 *val = wi->address;
 return(TRUE);
}

void delete_all_ws()
{
 watch_item *wi = first_wi;

 while(first_wi)
    {
     wi = first_wi;
     first_wi=first_wi->next;
     free(wi->name);
     free(wi);
    }
 first_wi = NULL;
 next_wi = NULL;
}

// restituisce il numero di simboli in tabella
int get_n_ws()
{
 int count =0;
 watch_item *wi=first_wi;

 while(wi)
    {
     wi=wi->next;
     count++;
    }
 return(count);
}

// restituisce il simbolo puntato da 'next_wi' in *s 
char *get_next_watch_item()
{
 char *result;

 if(next_wi)
   {
    result = next_wi->name;
    if( !(next_wi=next_wi->next) )
        next_wi = first_wi;
    return(result);
   }
 else
    return(NULL);
}