/* Da dove leggiamo il testo da assemblare?
   Qui c'e' tutto quello che riguarda l'allocazione di
   buffer di testo, la lettura da TEditWindow->GetText(),
   eccetera eccetera.

   -   open_buffer() controlla se il file e' gia' stato letto.
       Se non e' ancora stato letto alloca memoria per l'intero file,
       lo copia tutto nel buffer (eventualmente leggendolo da una finestra
       dell'editor) e mette in uno stack il puntatore al buffer.
       Se esiste gia' il buffer relativo al file, crea solo un nuovo
       puntatore nello stack. Si usa al posto di fopen(). 

   -   close_buffer() distrugge l'ultimo puntatore nello stack.
       Si usa al posto di fclose().
        
   -   buffer_getc() si usa per leggere testo dal buffer corrente.

   -   destroy_all_buffers distrugge tutti i puntatori dello stack
       (se per errore ne sono rimasti) e cancella tutti i buffers
       di testo.

   in aggiunta, per le macro:

   -   char *GetBufferPointer() restituisce il puntatore al prossimo
       carattere del buffer corrente.

   WARNING:
      NON FUNZIONA NIENTE con i files > 32k!!!!!!

*/

#include <stdlib.h>
#include <stdio.h>
#include <dir.h>
#include <string.h>
#include <windows.h>
#include "usermsg.h"

/* la lunghezza massima e' 32k */
#define MAX_LEN 32768

extern HWND FinestraPrincipale;

typedef struct t_textbuffer {
			     char buffer[MAX_LEN];
			     char filename[MAXPATH];
			     unsigned int length;
                             struct t_textbuffer *prev;
			    } textbuffer;

typedef struct t_bufferstack {
                              textbuffer *ptbuffer;
			      unsigned int pos;
			      struct t_bufferstack *prev;
                             }  bufferstack;

textbuffer *last_textbuffer = NULL;
bufferstack *current_buffer = NULL;

unsigned int buffer_getc(char *buf)
{
 if (current_buffer != NULL)
      if (current_buffer->pos < current_buffer->ptbuffer->length)
         {         
	  *buf = current_buffer->ptbuffer->buffer[current_buffer->pos++];
          return(1);
         }

 return(0);
}

int open_buffer(char *name)
{
 FILE *p;
 HWND handle;
 textbuffer *pt = last_textbuffer;
 bufferstack *newpt;
 char is_equal = 0;    /* 0 = FALSE per default. */

 while(pt)  /* Cerca se il file e' gia' stato bufferizzato */
    {
     if ( (is_equal = !strcmp(name, pt->filename))!=0 )
       break;
     else
       pt=pt->prev;
    }

 /* Se il file non e' stato ancora bufferizzato, crea anche il buffer */
 if (!is_equal)
    {
     handle = (HWND)SendMessage(FinestraPrincipale, WM_USER + FIND_FILE, 0, (LPARAM)name );
     if(!handle)  /* leggo da un file */
       {
        p = fopen (name, "r");
        if(p)
    	   {
            pt = malloc(sizeof(textbuffer));
	    pt->length =fread(pt->buffer, 1, MAX_LEN, p);
	    pt->prev = last_textbuffer;
	    strcpy(pt->filename, name);
	    last_textbuffer = pt;
            fclose(p);
	   }
        else
          return(NULL);  /* File non trovato; */
       }
     else          /* leggo da una finestra */
       {
	    pt = malloc(sizeof(textbuffer));
	    pt->length = SendMessage(handle, WM_GETTEXT, MAX_LEN, (LPARAM)(pt->buffer));
	    pt->prev = last_textbuffer;
	    strcpy(pt->filename, name);
	    last_textbuffer = pt;
	/* qui avevo messo un   fclose(p);  che faceva piantare tutto... */
       }
    }

 /* Se il file e' gia' stato bufferizzato, crea solo un nuovo puntatore */
 newpt = malloc(sizeof(bufferstack));
 newpt->prev = current_buffer;
 newpt->ptbuffer = pt;
 newpt->pos = 0;
 current_buffer = newpt;

 return(1);  /* OK */
}

/* Chiude l'ultimo buffer (senza liberare la memoria) */
void close_buffer()
{
 bufferstack *p;

 p=current_buffer;
 current_buffer=current_buffer->prev;
 free(p);
}

/* Spacca tutto. */
void destroy_all_buffers()
{
 textbuffer *pt;
 bufferstack *pts;

 while(current_buffer)
   {
    pts=current_buffer;
    current_buffer=current_buffer->prev;
    free(pts);
   }

 while(last_textbuffer)
   {
    pt=last_textbuffer;
    last_textbuffer = last_textbuffer->prev;
    free(pt);
   }

}

char *GetBufferPointer()
{
 if (current_buffer != NULL)
      return( current_buffer->ptbuffer->buffer + current_buffer->pos);
 else
      return("\0");
}
