/* Classe TWatchWindow */
#include "watch.h"
#include "defines.h"

#include <stdio.h>
#include <string.h>

#define max(a,b)(((a) >(b)) ?(a) :(b))
#define min(a,b)(((a) <(b)) ?(a) :(b))
#define abs(a)  (((a) >= 0)  ?(a) :-(a))

#define ID_WATCHLIST 1000
#define LOC_WIDTH 70
#define RES_WIDTH 40

// RAM per guardare le variabili
extern unsigned char huge RAM[65536];

//funzione che valuta l'espressione
extern "C" int exp_eval(char *, unsigned int *);

//funzioni per riempire le liste dei simboli
extern "C" int get_n_ws();
extern "C" char *get_next_watch_item();

// Dialog "Add expression... "

TAddDialog::TAddDialog(PTWindowsObject AParent, char *buf)
      :TDialog(AParent, ID_ADDWATCH)
{
 Destination = buf;
 Destination[0] = '\0';
}

void TAddDialog::SetupWindow()
{
 int nsimboli, i;
 char *pt;

 nsimboli = get_n_ws();
 for(i=1; i<=nsimboli; i++)
     {
      pt = get_next_watch_item();
      if (pt)
        SendDlgItemMessage(HWindow, ID_SYMBOLWATCH, LB_ADDSTRING, 0, (LPARAM)pt);
     }
}

void TAddDialog::WMCommand(RTMessage msg)
{
 char buf[255];
 int n;
 if(msg.WP.Lo)
    // se il messaggio arriva da un controllo:
    switch(msg.WParam)
       {
	case ID_INSERTEXPR:    // Inserisce il simbolo
	   n = SendDlgItemMsg(ID_SYMBOLWATCH, LB_GETCURSEL, 0, 0L);
	   SendDlgItemMsg(ID_SYMBOLWATCH, LB_GETTEXT, (WPARAM)n, (LPARAM)buf );
	   SendDlgItemMsg(ID_WATCHEXPR, EM_REPLACESEL, 0, (LPARAM)buf );
	   break;

	case IDOK:             // aggiunge l'espressione ed esce
	   SendDlgItemMsg(ID_WATCHEXPR, WM_GETTEXT, 254, (LPARAM)Destination);
	   break;

	case IDCANCEL:         // Esce senza aggiungere nulla
           break;

	case ID_SYMBOLWATCH:
           if( msg.LP.Hi== LBN_DBLCLK)
                {
	         n = SendDlgItemMsg(ID_SYMBOLWATCH, LB_GETCURSEL, 0, 0L);
	         SendDlgItemMsg(ID_SYMBOLWATCH, LB_GETTEXT, (WPARAM)n, (LPARAM)buf );
		 SendDlgItemMsg(ID_WATCHEXPR, EM_REPLACESEL, 0, (LPARAM)buf );
		}
           break;
       }

 TDialog::WMCommand(msg); // Risposte standard a OK e Cancel
}

void TAddDialog::CloseWindow(int n)
{
 TDialog::CloseWindow(n);
}

//--------------------------------------------------------------------
//Watch list
TWatchList::TWatchList(PTWindowsObject AParent, int AnId, int X, int Y,
		           int W, int H, PTModule AModule = NULL)
	     : TListBox(AParent, AnId, X, Y, W, H, AModule)
{
  Attr.Style = WS_CHILD | WS_VISIBLE | WS_GROUP | WS_TABSTOP | WS_VSCROLL
	      | LBS_NOTIFY | LBS_NOINTEGRALHEIGHT |LBS_USETABSTOPS ;
}


//----------------------------------------------------------
// Watch window

TWatchWindow::TWatchWindow(PTWindowsObject AParent, TWatchWindow **Me)
      : TWindow(AParent, "Watch")
{
 Myself = Me;
 Elenco = new TWatchList(this, ID_WATCHLIST, 1,1,1,1);
 first_watch_item = NULL;
}

TWatchWindow::~TWatchWindow()
{
 ClearWatchList();
 *Myself = NULL;
 delete Elenco;

}

void TWatchWindow::SetupWindow()
{
 HFONT f;

 TWindow::SetupWindow();
 f = CreateFont(-12, 0, 0, 0, FW_BOLD, 0, 0, 0, 0, 0, 0, 0, 0, "MS Sans Serif");
 SendMessage(Elenco->HWindow, WM_SETFONT, (WPARAM)f, (LPARAM) MAKELONG((WORD) TRUE, 0) );
}

LPSTR TWatchWindow::GetClassName()
{
 return("WatchWindow");
}

// Assegna l'icona della finestra
void TWatchWindow::GetWindowClass(WNDCLASS& AWndClass)
{
  TWindow::GetWindowClass(AWndClass);
  AWndClass.hIcon = LoadIcon(GetApplication()->hInstance, "ICONA_WATCH");
}

// Modifica il menu della finestra MDI
void TWatchWindow::ActivationResponse(WORD Activated, BOOL IsIconified)
{
 HWND handle;
 HMENU hm;

 TWindow::ActivationResponse(Activated, IsIconified);
 handle = GetApplication()->MainWindow->HWindow;
 hm = GetMenu(handle);
 if(Activated)
     {
//menu File
      EnableMenuItem(hm, CM_FILESAVE,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_FILESAVEAS,  MF_BYCOMMAND | MF_GRAYED);
//menu Edit
      EnableMenuItem(hm, CM_EDITUNDO,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCUT,     MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCOPY,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITPASTE,   MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITDELETE,  MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCLEAR,   MF_BYCOMMAND | MF_GRAYED);
//menu Search
      EnableMenuItem(hm, CM_EDITFIND,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITREPLACE, MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITFINDNEXT,MF_BYCOMMAND | MF_GRAYED);
//menu Assembler
      EnableMenuItem(hm, MENU_ASSEMBLER_ASSEMBLE, MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, MENU_ASSEMBLER_BUILDALL, MF_BYCOMMAND | MF_GRAYED);
     }
 else
     {
//menu File
      EnableMenuItem(hm, CM_FILESAVE,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_FILESAVEAS,  MF_BYCOMMAND | MF_ENABLED);
//menu Edit
      EnableMenuItem(hm, CM_EDITUNDO,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCUT,     MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCOPY,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITPASTE,   MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITDELETE,  MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCLEAR,   MF_BYCOMMAND | MF_ENABLED);
//menu Search
      EnableMenuItem(hm, CM_EDITFIND,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITREPLACE, MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITFINDNEXT,MF_BYCOMMAND | MF_ENABLED);
//menu Assembler
      EnableMenuItem(hm, MENU_ASSEMBLER_ASSEMBLE, MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, MENU_ASSEMBLER_BUILDALL, MF_BYCOMMAND | MF_ENABLED);
     }
 DrawMenuBar(handle);
}

void TWatchWindow::WMSize(RTMessage msg)
{
 int xd[2];

 TWindow::WMSize(msg);
 if (msg.WParam != SIZE_MINIMIZED)
     {
      xd[0] = ((msg.LP.Lo-RES_WIDTH-LOC_WIDTH)*4)/LOWORD(GetDialogBaseUnits())+1;
      xd[1] = ((msg.LP.Lo-RES_WIDTH)*4)/LOWORD(GetDialogBaseUnits())+1;
      SendMessage(Elenco->HWindow, LB_SETTABSTOPS, (WPARAM)2, (LPARAM)xd);
      MoveWindow(Elenco->HWindow, 0, 0, msg.LP.Lo, msg.LP.Hi, TRUE);
     }
}


void TWatchWindow::AddWatch(RTMessage)
{
 char buffer[255];

 do
   {
    GetApplication()->ExecDialog(new TAddDialog(this, buffer));
    if(buffer[0])
      {
       AddExpression(buffer);
       RicalcolaTutto();
      }
   }
 while(buffer[0]);
}

void TWatchWindow::RemoveWatch(RTMessage)
{
 RemoveExpression();
 RicalcolaTutto();
}

void TWatchWindow::RemoveAllWatches(RTMessage)
{
 ClearWatchList();
}

void TWatchWindow::AddExpression(char *expr)
{
 watch_item *pt;

 pt = (t_watch_item *)malloc(sizeof(watch_item));
 pt->expr = (char *)malloc(strlen(expr)+1);
 strcpy(pt->expr, expr);
 pt->next= first_watch_item;
 first_watch_item = pt;
}

void TWatchWindow::RemoveExpression()
{
 int i,n;
 watch_item *pt = first_watch_item;

 n = Elenco->GetSelIndex();
 if(n==0)
    { //primo elemento
     first_watch_item = first_watch_item->next;
     free(pt->expr);
     free(pt);
    }
 else
    {
     watch_item *tmp;
     n--;
     i=0;
     while(pt && (i<n))
       {
	pt=pt->next;
        i++;
       }
     if(!pt)
        return;
     tmp = pt->next;
     if(tmp)
        {
	 pt->next = tmp->next;
         free(tmp);
        }
    }
}

void TWatchWindow::ClearWatchList()
{
 watch_item *pt;

 while(first_watch_item)
    {
     pt = first_watch_item;
     first_watch_item = first_watch_item->next;
     free(pt->expr);
     free(pt);
    }
}

void TWatchWindow::WMCommand(RTMessage msg)
{
 /*
 if(msg.WP.Lo)
    // se il messaggio arriva da un controllo:
    switch(msg.LP.Hi)
       {
	case LBN_SELCHANGE:
	   switch(msg.WParam)
              {
	       case ID_WATCHLIST:
		  break;
	       }
	   break;
       }
 */
 TWindow::WMCommand(msg);
}

void TWatchWindow::RicalcolaTutto()
{
 char buffer[255];
 unsigned int val;
 watch_item *pt = first_watch_item;

 Elenco->ClearList();

 while(pt)
    {
     if ( exp_eval(pt->expr, &val) )
	  {
	     //NOTA: non uso, per ora, pt->loc perche' suppongo
	     // che le espressioni siano variabili 
	   sprintf(buffer, "%s\t%04X\t%02X",pt->expr, val, RAM[val]);
	   Elenco->AddString(buffer);
	  }
     else
	  {
	   sprintf(buffer, "%s\t????\t??", pt->expr);
	   Elenco->AddString(buffer);
	  }
     pt=pt->next;
    }
 
}