/* Classe TWatchWindow */
#include "watch.h"
#include "defines.h"

#include <stdio.h>
#include <string.h>

#define max(a,b)(((a) >(b)) ?(a) :(b))
#define min(a,b)(((a) <(b)) ?(a) :(b))
#define abs(a)  (((a) >= 0)  ?(a) :-(a))

#define ID_WATCHLIST 1000
#define ID_WATCHLOCATIONS 1100
#define ID_WATCHRESULTS 1200
#define LOC_WIDTH 70
#define RES_WIDTH 40

// RAM per guardare le variabili
extern unsigned char huge RAM[65536];

//funzione che valuta l'espressione
extern "C" int exp_eval(char *, unsigned int *);

//funzioni per riempire le liste dei simboli
extern "C" int get_n_ws();
extern "C" char *get_next_watch_item();

// Dialog "Add expression... "

TAddDialog::TAddDialog(PTWindowsObject AParent, char *buf)
      :TDialog(AParent, ID_ADDWATCH)
{
 Destination = buf;
 Destination[0] = '\0';
}

void TAddDialog::SetupWindow()
{
 int nsimboli, i;
 char *pt;

 nsimboli = get_n_ws();
 for(i=1; i<=nsimboli; i++)
     {
      pt = get_next_watch_item();
      if (pt)
        SendDlgItemMessage(HWindow, ID_SYMBOLWATCH, LB_ADDSTRING, 0, (LPARAM)pt);
     }
}

void TAddDialog::WMCommand(RTMessage msg)
{
 if(msg.WP.Lo)
    // se il messaggio arriva da un controllo:
    switch(msg.WParam)
       {
	case ID_INSERTEXPR:    // Inserisce il simbolo
	   break;

	case IDOK:             // aggiunge l'espressione ed esce
	   SendDlgItemMsg(ID_WATCHEXPR, WM_GETTEXT, 254, (LPARAM)Destination);
	   break;

	case IDCANCEL:         // Esce senza aggiungere nulla
           break;
       }

 TDialog::WMCommand(msg); // Risposte standard a OK e Cancel
}

void TAddDialog::CloseWindow(int n)
{
 TDialog::CloseWindow(n);
}

//--------------------------------------------------------------------
//Watch list
TWatchList::TWatchList(PTWindowsObject AParent, int AnId, int X, int Y,
		           int W, int H, BOOL HasSB, PTModule AModule = NULL)
	     : TListBox(AParent, AnId, X, Y, W, H, AModule)
{
  Attr.Style = WS_CHILD | WS_VISIBLE | WS_GROUP | WS_TABSTOP
	      | LBS_NOTIFY | LBS_NOINTEGRALHEIGHT ;
  if(HasSB)
      Attr.Style |= WS_VSCROLL;
}


//----------------------------------------------------------
// Watch window

TWatchWindow::TWatchWindow(PTWindowsObject AParent, TWatchWindow **Me)
      : TWindow(AParent, "Watch")
{
 Myself = Me;
 Elenco = new TWatchList(this, ID_WATCHLIST, 1,1,1,1, FALSE);
 Locations = new TWatchList(this, ID_WATCHLOCATIONS, 1,1,1,1, FALSE);
 Results = new TWatchList(this, ID_WATCHRESULTS, 1,1,1,1, TRUE);
}

TWatchWindow::~TWatchWindow()
{
 *Myself = NULL;
 delete Elenco;
 delete Locations;
 delete Results;
}

void TWatchWindow::SetupWindow()
{
 HFONT f;

 TWindow::SetupWindow();
 f = CreateFont(-12, 0, 0, 0, FW_BOLD, 0, 0, 0, 0, 0, 0, 0, 0, "MS Sans Serif");
 SendMessage(Elenco->HWindow, WM_SETFONT, (WPARAM)f, (LPARAM) MAKELONG((WORD) TRUE, 0) );
 SendMessage(Locations->HWindow, WM_SETFONT, (WPARAM)f, (LPARAM) MAKELONG((WORD) TRUE, 0) );
 SendMessage(Results->HWindow, WM_SETFONT, (WPARAM)f, (LPARAM) MAKELONG((WORD) TRUE, 0) );
}

LPSTR TWatchWindow::GetClassName()
{
 return("WatchWindow");
}

// Assegna l'icona della finestra
void TWatchWindow::GetWindowClass(WNDCLASS& AWndClass)
{
  TWindow::GetWindowClass(AWndClass);
  AWndClass.hIcon = LoadIcon(GetApplication()->hInstance, "ICONA_WATCH");
}

// Modifica il menu della finestra MDI
void TWatchWindow::ActivationResponse(WORD Activated, BOOL IsIconified)
{
 HWND handle;
 HMENU hm;

 TWindow::ActivationResponse(Activated, IsIconified);
 handle = GetApplication()->MainWindow->HWindow;
 hm = GetMenu(handle);
 if(Activated)
     {
//menu File
      EnableMenuItem(hm, CM_FILESAVE,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_FILESAVEAS,  MF_BYCOMMAND | MF_GRAYED);
//menu Edit
      EnableMenuItem(hm, CM_EDITUNDO,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCUT,     MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCOPY,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITPASTE,   MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITDELETE,  MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITCLEAR,   MF_BYCOMMAND | MF_GRAYED);
//menu Search
      EnableMenuItem(hm, CM_EDITFIND,    MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITREPLACE, MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, CM_EDITFINDNEXT,MF_BYCOMMAND | MF_GRAYED);
//menu Assembler
      EnableMenuItem(hm, MENU_ASSEMBLER_ASSEMBLE, MF_BYCOMMAND | MF_GRAYED);
      EnableMenuItem(hm, MENU_ASSEMBLER_BUILDALL, MF_BYCOMMAND | MF_GRAYED);
     }
 else
     {
//menu File
      EnableMenuItem(hm, CM_FILESAVE,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_FILESAVEAS,  MF_BYCOMMAND | MF_ENABLED);
//menu Edit
      EnableMenuItem(hm, CM_EDITUNDO,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCUT,     MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCOPY,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITPASTE,   MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITDELETE,  MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITCLEAR,   MF_BYCOMMAND | MF_ENABLED);
//menu Search
      EnableMenuItem(hm, CM_EDITFIND,    MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITREPLACE, MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, CM_EDITFINDNEXT,MF_BYCOMMAND | MF_ENABLED);
//menu Assembler
      EnableMenuItem(hm, MENU_ASSEMBLER_ASSEMBLE, MF_BYCOMMAND | MF_ENABLED);
      EnableMenuItem(hm, MENU_ASSEMBLER_BUILDALL, MF_BYCOMMAND | MF_ENABLED);
     }
 DrawMenuBar(handle);
}

void TWatchWindow::WMSize(RTMessage msg)
{
 int xd1, xd2;

 TWindow::WMSize(msg);
 if (msg.WParam != SIZE_MINIMIZED)
     {
      xd1 = msg.LP.Lo-RES_WIDTH-LOC_WIDTH;
      xd2 = msg.LP.Lo-RES_WIDTH;
      MoveWindow(Elenco->HWindow, 0, 0, xd1, msg.LP.Hi, TRUE);
      MoveWindow(Locations->HWindow, xd1, 0, LOC_WIDTH, msg.LP.Hi, TRUE);
      MoveWindow(Results->HWindow, xd2, 0, RES_WIDTH, msg.LP.Hi, TRUE);
     }
}


void TWatchWindow::AddWatch(RTMessage)
{
 char buffer[255];

 do
   {
    GetApplication()->ExecDialog(new TAddDialog(this, buffer));
    if(buffer[0])
      {
       AddExpression(buffer);
       RicalcolaTutto();
      }
   }
 while(buffer[0]);
}

void TWatchWindow::RemoveWatch(RTMessage)
{
 RemoveExpression();
}

void TWatchWindow::RemoveAllWatches(RTMessage)
{
 ClearWatchList();
}

void TWatchWindow::AddExpression(char *expr)
{
 Elenco->AddString(expr);
}

void TWatchWindow::RemoveExpression()
{
 Elenco->DeleteString(Elenco->GetSelIndex());
}

void TWatchWindow::ClearWatchList()
{
 Elenco->ClearList();
}

void TWatchWindow::WMCommand(RTMessage msg)
{
 if(msg.WP.Lo)
    // se il messaggio arriva da un controllo:
    switch(msg.LP.Hi)
       {
	case LBN_SELCHANGE:
	   switch(msg.WParam)
              {
	       case ID_WATCHLIST:
		  Locations->SetSelIndex(Elenco->GetSelIndex() );
		  Results->SetSelIndex(Elenco->GetSelIndex() );
                  break;
	       case ID_WATCHLOCATIONS:
		  Elenco->SetSelIndex(Locations->GetSelIndex() );
		  Results->SetSelIndex(Locations->GetSelIndex() );
                  break;
	       case ID_WATCHRESULTS:
		  Elenco->SetSelIndex(Results->GetSelIndex() );
		  Locations->SetSelIndex(Results->GetSelIndex() );
		  break;
	       }
	   break;
       }

 TWindow::WMCommand(msg);
}

void TWatchWindow::RicalcolaTutto()
{
 char buffer[255];
 char vstring[8];
 unsigned int val;
 int i, n;

 Results->ClearList();
 Locations->ClearList();

 n=Elenco->GetCount();
 for(i=0; i<n; i++)
    {
     Elenco->GetString(buffer, i);
     if ( exp_eval(buffer, &val) )
	  {
	   sprintf(vstring, "%04X", val);
           Locations->AddString(vstring);
	   sprintf(vstring, "%02X", RAM[val]);
           Results->AddString(vstring);
	  }
     else
	   {
	    Locations->AddString("???");
	    Results->AddString("???");
           }
    }
 
}