/*

Menu handling

*/


#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include "menu.h"
#include "screen.h"


unsigned char menu_bc =  1; /* border color */
unsigned char menu_tc =  7; /* title color */
unsigned char menu_nc = 13; /* normal item color */
unsigned char menu_ic =  5; /* inactive item color */


char DrawMenu(Menu *menu);
void DrawItem(Item *item, unsigned char xpos, unsigned char ypos, unsigned char current);


void SetMenuColors(unsigned char bc, unsigned char tc, unsigned char nc, unsigned char ic) {
  menu_bc = bc;
  menu_tc = tc;
  menu_nc = nc;
  menu_ic = ic;
}


Menu *NewMenu(void) {
  Menu *menu;

  if ((menu = malloc(sizeof(*menu))) == NULL) {
    return(NULL);
  }
  menu->visible = 0;
  menu->numitems = 0;
  menu->current = 0;
  menu->xpos = 10;
  menu->ypos = 2;
  menu->width = 0;
  menu->height = 0;
  menu->title = NULL;
  menu->firstitem = NULL;
  return(menu);
}


void SetMenuTitle(Menu *menu, unsigned char *title) {
  unsigned char l;

  menu->title = title;
  if ((l = strlen(title)) > menu->width) {
    menu->width = l;
  }
}


char AddMenuItem(Menu *menu, unsigned char id, ItemState state, unsigned char *name) {
  Item *item;
  int l;

  if (menu->numitems == 0) {
    if ((menu->firstitem = malloc(sizeof(*item))) == NULL) {
      return(-1);
    }
    item = menu->firstitem;
    item->prev = NULL;
  } else {
    item = menu->firstitem;
    while (item->next) {
      item = item->next;
    }
    if ((item->next = malloc(sizeof(*item))) == NULL) {
      return(-1);
    }
    item->next->prev = item;
    item = item->next;
  }
  item->next = NULL;
  item->id = id;
  item->name = name;
  item->state = state;
  ++(menu->numitems);
  if ((l = strlen(name)) > menu->width) {
    menu->width = l;
  }
  if (menu->numitems > menu->height) {
    menu->height = menu->numitems;
  }
  return(0);
}


char DelMenuItem(Menu *menu, unsigned char id) {
  Item *item;

  if (menu->numitems) {
    item = menu->firstitem;
    while (item) {
      if (item->id == id) {
	if (menu->firstitem == item) {
	  menu->firstitem = item->next;
	  item->next->prev = NULL;
	  free(item);
	} else {
	  item->prev->next = item->next;
	  item->next->prev = item->prev;
	  free(item);
	}
	--(menu->numitems);
	return(0);
      }
      item = item->next;
    }
  }
  return(-1);
}


char SetItemState(Menu *menu, unsigned char id, ItemState state) {
  Item *item;

  if (menu->numitems) {
    item = menu->firstitem;
    while (item) {
      if (item->id == id) {
	item->state = state;
	return(0);
      }
      item = item->next;
    }
  }
  return(-1);
}


void ShowMenu(Menu *menu) {
  if (menu->visible == 0) {
    menu->visible = 1;
    DrawMenu(menu);
  }
}


void HideMenu(Menu *menu) {
  if (menu->visible) {
    menu->visible = 0;
    DrawMenu(menu);
  }
}


Item *FindItemNumber(Menu *menu, unsigned char itemnum) {
  Item *item;

  item = menu->firstitem;
  while (itemnum) {
    if (item->next) {
      item = item->next;
      --itemnum;
    } else {
      return(NULL);
    }
  }
  return(item);
}


Item *FindItem(Menu *menu, unsigned char id) {
  Item *item;

  item = menu->firstitem;
  while (item->id != id) {
    if (item->next) {
      item = item->next;
    } else {
      return(NULL);
    }
  }
  return(item);
}


char SelectMenu(Menu *menu) {
  unsigned char c;
  Item *item;

  if ((item = FindItemNumber(menu, menu->current)) == NULL) {
    return(-1);
  }
  if (menu->visible == 0) {
    return(-1);
  }
  while(c = cgetc()) {
    switch (c) {
    case 10:
    case 13:
      if (item->state == ISTATE_NORMAL) {
	return(item->id);
      }
      break;
    case CH_CURS_DOWN:
      if (menu->current + 1 < menu->numitems) {
	DrawItem(item, menu->xpos + 1, menu->ypos + menu->current + 3, 0);
	item = item->next;
	++(menu->current);
	DrawItem(item, menu->xpos + 1, menu->ypos + menu->current + 3, 1);
      }
      break;
    case CH_CURS_UP:
      if (menu->current > 0) {
	DrawItem(item, menu->xpos + 1, menu->ypos + menu->current + 3, 0);
	item = item->prev;
	--(menu->current);
	DrawItem(item, menu->xpos + 1, menu->ypos + menu->current + 3, 1);
      }
      break;
    }
  }
}


void DestroyMenu(Menu *menu) {
  Item *item, *next;

  if (menu->numitems) {
    item = menu->firstitem;
    while (item) {
      next = item->next;
      free(item);
      item = next;
    }
  }
  free(menu);
}


void DrawItem(Item *item, unsigned char xpos, unsigned char ypos, unsigned char current) {
  unsigned char c;
  gotoxy(xpos, ypos);
  if (current) {
    revers(1);
  }
  if (item->state == ISTATE_NORMAL) {
    c = textcolor(menu_nc);
    cprintf(item->name);
  } else {
    c = textcolor(menu_ic);
    cprintf(item->name);
  }
  textcolor(c);
  revers(0);
}


char DrawMenu(Menu *menu) {
  unsigned char l = 0;
  Item *item;

  if (menu->numitems == 0) {
    return(-1);
  } else {
    if (menu->title == NULL) {
      return(-1);
    } else {
      if (menu->visible) {
	textcolor(menu_bc);
	DrawBox(menu->width+2,menu->height+4,menu->xpos,menu->ypos);
	gotoxy(menu->xpos + 1, menu->ypos + 1);
	textcolor(menu_tc);
	cprintf(menu->title);
	item = menu->firstitem;
	while (item) {
	  DrawItem(item, menu->xpos + 1, menu->ypos + l + 3, menu->current == l ? 1 : 0);
	  item = item->next;
	  ++l;
	}
      } else {
	ClearBox(menu->width+2,menu->height+4,menu->xpos,menu->ypos);
      }
    }
  }
}
