/*

Menu handling

*/


extern unsigned char menu_bc; /* border color */
extern unsigned char menu_tc; /* title color */
extern unsigned char menu_nc; /* normal item color */
extern unsigned char menu_ic; /* inactive item color */


typedef enum itemstates {
  ISTATE_NORMAL,
  ISTATE_INACTIVE
} ItemState;

typedef struct item {
  struct item *next;
  struct item *prev;
  unsigned char id;
  unsigned char *name;
  ItemState state;
} Item;

typedef struct menu {
  unsigned char visible;
  unsigned char numitems;
  unsigned char current;
  unsigned char xpos;
  unsigned char ypos;
  unsigned char width;
  unsigned char height;
  unsigned char *title;
  Item *firstitem;
} Menu;


/* Set menu colors */
void SetMenuColors(unsigned char bc, unsigned char tc, unsigned char nc, unsigned char ic);

/* Initialize an empty menu. */
Menu *NewMenu(void);

/* Deallocate menu resources. */
void DestroyMenu(Menu *menu);

/* Set the title. */
void SetMenuTitle(Menu *menu, unsigned char *title);

/* Add an item to the menu. id is the value that SelectMenu will return. */
char AddMenuItem(Menu *menu, unsigned char id, ItemState state, unsigned char *name);

/* Delete an item from the menu. */
char DelMenuItem(Menu *menu, unsigned char id);

/* Set item state to normal or inactive. */
char SetItemState(Menu *menu, unsigned char id, ItemState state);

/* Display the menu on the screen. */
void ShowMenu(Menu *menu);

/* Erase the menu from the screen. */
void HideMenu(Menu *menu);

/* Let the user choose a menu item. ShowMenu() first. */
char SelectMenu(Menu *menu);
